/* --------------------------------------------------------------------------
$Id: hlhdf.c,v 1.8 2004/09/21 10:55:55 ahenja Exp $
Program:        

Description:    Common functions for working with an HDF5 file.

Author(s):      Anders Henja

Copyright:	Swedish Meteorological and Hydrological Institute, 2000

History:
$Log: hlhdf.c,v $
Revision 1.8  2004/09/21 10:55:55  ahenja
Fixed weird memory problem, needed H5Tclose(f_memb) in getFixedType
Only added for versions > 1.4

Revision 1.7  2004/09/20 09:58:16  ahenja
Fixed strange bug!?!?!?!
:wq!

Revision 1.6  2004/07/01 09:28:16  ahenja
Updated with support for new compression and new loghandling

Revision 1.5  2004/03/29 14:58:21  ahenja
Updated with support for turning on/off outputs from
both HLHDF and HDF5 all depending on needs, also added
so that it is possible to reroute error messages by
writing client specific functions that handles the
various messages.

Revision 1.4  2004/03/18 11:53:49  ahenja
Modified with support for szip, also changed so that it is possible
to specify HL_FileCreationProperty when creating a file. Also changed
how the compression handling is working. Added a nicer way to
perform debugging and error handling. Now it is possible to define
an own error handler and hook it into the system, if this is done
all error messages will be sent to the given function.

Revision 1.3  2002/12/19 13:29:14  ahenja
Added some typecastings when mallocing memory.

Revision 1.2  2002/08/06 10:29:16  ahenja
Corrected bug in write_hlhdf.c function addAttributeToCompoundType, an illegal
cast was performed for HDF5 Versions >= 1.4.x.

Revision 1.1.1.1  2001/12/14 12:02:39  ahenja
First CVS version of hlhdf

Revision 1.17  2001/11/13 15:18:44  adybbroe
*** empty log message ***

Revision 1.16  2001/11/12 10:25:24  adybbroe
corrected for new type H5T_ARRAY.

Revision 1.15  2001/11/08 08:54:35  ahenja
Removed comments, etc. seems like Adam has solved all problems related to HDF5.4.2, I don't know if it works with HDF5.4.1

Revision 1.14  2001/11/08 08:37:31  adybbroe
Forced checkin since ahenja need access to file.

Revision 1.13  2001/04/26 14:17:06  adybbroe
Solved for version 1.4 incompatibility.

Revision 1.12  2000/04/11 10:07:50  ahenja
Added function for checking wether a file is hdf5 format or not.

Revision 1.11  2000/03/22 08:35:25  ahenja
Changed description.

Revision 1.10  2000/03/22 08:31:25  ahenja
Updated with spelling corrections and header information.

-----------------------------------------------------------------------------*/
#include "hlhdf.h"
#include "hlhdf_debug.h"

/*For internal use*/
static int errorReportingOn=1;
static void *edata;
static hid_t (*errorFunction)(void*);

int _debug_hdf;

/************************************************
 * disableErrorReporting
 ***********************************************/
void disableErrorReporting()
{
   /*Disable error reporting*/
   if(errorReportingOn==1) {
      H5Eget_auto (&errorFunction, &edata);
      H5Eset_auto (NULL, NULL);
      errorReportingOn=0;
   }
}

/************************************************
 * enableErrorReporting
 ***********************************************/
void enableErrorReporting()
{
   if(errorReportingOn==0) {
      H5Eset_auto (errorFunction, edata);
      errorReportingOn=1;
   }
}

/************************************************
 * initHlHdf
 ***********************************************/
void initHlHdf()
{
   _debug_hdf=0;
   H5Eset_auto(HL_hdf5_debug_function,NULL); /* Force logging to always goto HLHDFs handler */
   HL_InitializeDebugger();
   HL_enableHdf5ErrorReporting();
   disableErrorReporting();
}

/************************************************
 * debugHlhdf
 ***********************************************/
void debugHlHdf(int flag)
{
   if(flag==0) {
      /*Don't debug anything*/
      _debug_hdf=0;
      disableErrorReporting();
   } else if(flag==1) {
      /*Only debug HLHDF stuff*/
      _debug_hdf=1;
      HL_setDebugLevel(HLHDF_DEBUG);
      disableErrorReporting();
   } else {
      /*Debug everything*/
      _debug_hdf=1;
      HL_setDebugLevel(HLHDF_DEBUG);
      enableErrorReporting();
   }
}
/************************************************
 * isHdf5File
 ***********************************************/
int isHdf5File(const char* filename)
{
   htri_t checkValue=H5Fis_hdf5(filename);
   HL_DEBUG0("isHdf5File");
   if(checkValue > 0)
      return 1;
   else
      return 0;
}

/************************************************
 * openHlHdfFile
 ***********************************************/
hid_t openHlHdfFile(const char* filename,const char* how)
{
   unsigned flags = H5F_ACC_RDWR;
   HL_DEBUG0("ENTER: openHlHdfFile");
   
   if(strcmp(how,"r") == 0) {
      flags = H5F_ACC_RDONLY;
   } else if(strcmp(how,"w")==0 || strcmp(how,"rw")==0 || strcmp(how,"wr")==0) {
      flags = H5F_ACC_RDWR;
   } else {
      HL_ERROR0("Illegal mode given when opening file, should be (r|w|rw)");
      return (hid_t)-1;
   }
   return H5Fopen(filename,flags,H5P_DEFAULT);
}

/************************************************
 * createHlHdfFileWithProperty
 ***********************************************/
hid_t createHlHdfFileWithProperty(const char* filename, HL_FileCreationProperty* property)
{
#if H5_VERS_MAJOR>1 || (H5_VERS_MAJOR<2 && H5_VERS_MINOR>=6 && H5_VERS_RELEASE>=1)
   /* HDF5 releases lower than 1.6.1 contains a bug that will cause the file to be
    * corrupt, unreadable if modifying the properties, so deactivate this feature
    * if the release is lower than 1.6.1 */
   hid_t propId = -1;
   hid_t fileId = -1;
   hid_t fileaccesspropertyId = -1;
   
   HL_DEBUG0("ENTER: createHlHdfFileWithProperty");
   HL_DEBUG0("Using default properties");
   if(property==NULL) {
      return H5Fcreate(filename,H5F_ACC_TRUNC,H5P_DEFAULT,H5P_DEFAULT);
   } else {
      HL_DEBUG0("Using specific properties");
      if((propId=H5Pcreate(H5P_FILE_CREATE))<0) {
	 HL_ERROR0("Failed to create the property");
	 goto fail;
      }
      HL_DEBUG1("Setting userblock property to %d",property->userblock);
      if(H5Pset_userblock(propId,property->userblock)<0) {
	 HL_ERROR0("Failed to set the userblock property");
	 goto fail;
      }
      HL_DEBUG2("Setting sizes to %d, %d",property->sizes.sizeof_addr,property->sizes.sizeof_size);
      if(H5Pset_sizes(propId,property->sizes.sizeof_addr,property->sizes.sizeof_size)<0) {
	 HL_ERROR0("Failed to set the sizes property");
	 goto fail;
      }
      HL_DEBUG2("Setting sym_k to %d, %d",property->sym_k.ik,property->sym_k.lk);
      if(H5Pset_sym_k(propId,property->sym_k.ik,property->sym_k.lk)<0) {
	 HL_ERROR0("Failed to set the sym_k property");
	 goto fail;
      }
      HL_DEBUG1("Setting istore_k to %d",property->istore_k);
      if(H5Pset_istore_k(propId,property->istore_k)<0) {
	 HL_ERROR0("Failed to set the istore_k property");
	 goto fail;
      }
      if(property->meta_block_size!=2048) {
	 if((fileaccesspropertyId=H5Pcreate(H5P_FILE_ACCESS))<0) {
	    HL_ERROR0("Failed to create the H5P_FILE_ACCESS property");
	    goto fail;
	 }
	 if(H5Pset_meta_block_size(fileaccesspropertyId,property->meta_block_size)<0) {
	    HL_ERROR0("Failed to set the meta block size");
	    goto fail;
	 }
	 fileId = H5Fcreate(filename,H5F_ACC_TRUNC,propId,fileaccesspropertyId);
	 H5Pclose(fileaccesspropertyId);
	 H5Pclose(propId);
	 return fileId;
      } else {
	 fileId = H5Fcreate(filename,H5F_ACC_TRUNC,propId,H5P_DEFAULT);
	 H5Pclose(propId);
	 return fileId;
      }
   }
 fail:
   if(propId>=0)
      H5Pclose(propId);
   if(fileaccesspropertyId>=0)
      H5Pclose(fileaccesspropertyId);
   return -1;
#else
   HL_DEBUG0("ENTER: createHlHdfFileWithProperty");
   HL_DEBUG0("Using default properties");
   return H5Fcreate(filename,H5F_ACC_TRUNC,H5P_DEFAULT,H5P_DEFAULT);
#endif
   
}

/************************************************
 * createHlHdfFile
 ***********************************************/
hid_t createHlHdfFile(const char* filename)
{
   HL_DEBUG0("ENTER: createHlHdfFile");
   return createHlHdfFileWithProperty(filename,NULL);
}

/************************************************
 * createHlHdfFileCreationProperty
 ***********************************************/
HL_FileCreationProperty* createHlHdfFileCreationProperty()
{
   HL_FileCreationProperty* retv = NULL;
   hid_t theHid = -1;

   HL_DEBUG0("ENTER: createHlHdfFileCreationProperty");
   
   if((retv = (HL_FileCreationProperty*)malloc(sizeof(HL_FileCreationProperty)))==NULL) {
      HL_ERROR0("Failure when allocating memory for HL_FileCreationProperty");
      return NULL;
   }

   if((theHid = H5Pcreate(H5P_FILE_CREATE))<0) {
      HL_ERROR0("Failure when creating the property list");
      free(retv);
      return NULL;
   }
   
   /* Fetch default information */
   if(H5Pget_version(theHid, &retv->version.super, &retv->version.freelist,
		     &retv->version.stab, &retv->version.shhdr)<0) {
      HL_ERROR0("Failure while getting version for property");
      goto fail;
   }

   if(H5Pget_userblock(theHid,&retv->userblock)<0) {
      HL_ERROR0("Failure while getting the userblock for property");
      goto fail;
   }

   if(H5Pget_sizes(theHid,&retv->sizes.sizeof_addr,&retv->sizes.sizeof_size)<0) {
      HL_ERROR0("Failure while getting the sizes for property");
      goto fail;
   }

   if(H5Pget_sym_k(theHid,&retv->sym_k.ik, &retv->sym_k.lk)<0) {
      HL_ERROR0("Failure while getting the sym_k for property");
      goto fail;
   }

   if(H5Pget_istore_k(theHid,&retv->istore_k)<0) {
      HL_ERROR0("Failure while getting the istore_k for property");
      goto fail;
   }
   if(theHid>=0) {
      H5Pclose(theHid);
   }
   theHid=-1;
   if((theHid = H5Pcreate(H5P_FILE_ACCESS))<0) {
      HL_ERROR0("Failure when creating the file access property list");
      goto fail;
   }
   if(H5Pget_meta_block_size(theHid,&retv->meta_block_size)<0) {
      HL_ERROR0("Failure while getting the meta_block_size for property");
      goto fail;
   }
   if(theHid>=0) {
      H5Pclose(theHid);
   }
   theHid=-1;
      
   return retv;
 fail:
   if(theHid>=0) {
      H5Pclose(theHid);
   }
   freeHL_fileCreationProperty(retv);
   return NULL;
}

/************************************************
 * freeHL_fileCreationProperty
 ***********************************************/
void freeHL_fileCreationProperty(HL_FileCreationProperty* prop)
{
   HL_DEBUG0("ENTER: freeHL_fileCreationProperty");
   if(prop==NULL) {
      return;
   }
   free(prop);
}

/************************************************
 * closeHlhdfFile
 ***********************************************/
herr_t closeHlHdfFile(hid_t file_id)
{
   HL_DEBUG0("ENTER: closeHlHdfFile");
   return H5Fclose(file_id);
}

/************************************************
 * getFixedType
 ***********************************************/
#ifdef H5_VERS_MINOR
/*#if H5_VERS_MINOR>=4 */
#if H5_VERS_MAJOR>1 || (H5_VERS_MAJOR<2 && H5_VERS_MINOR>=4)
hid_t getFixedType(hid_t type)
{
   size_t size;
   hid_t mtype=-1;
   H5T_str_t strpad;
   hid_t* memb=NULL;
   char** name=NULL;
   int nmembs=0,i,j,*ndims=NULL;
   size_t offset,nelmts;
   hsize_t *dims141=NULL;
   hid_t f_memb=-1;
   hid_t member_type;
   hid_t tmpt=-1;
   
   HL_SPEWDEBUG0("ENTER: getFixedType, version > 1.4");
   
   size=H5Tget_size(type);

   switch(H5Tget_class(type)) {
   case H5T_INTEGER:
      HL_SPEWDEBUG0("This is of type H5T_INTEGER");
      if(size<=sizeof(char))
	 mtype=H5Tcopy(H5T_NATIVE_SCHAR);
      else if(size<=sizeof(short))
	 mtype=H5Tcopy(H5T_NATIVE_SHORT);
      else if(size<=sizeof(int))
	 mtype=H5Tcopy(H5T_NATIVE_INT);
      else if(size<=sizeof(long))
	 mtype=H5Tcopy(H5T_NATIVE_LONG);
      else
	 mtype=H5Tcopy(H5T_NATIVE_LLONG);
      H5Tset_sign(mtype,H5Tget_sign(type));
      break;
   case H5T_FLOAT:
      HL_SPEWDEBUG0("This is of type H5T_FLOAT");
      if(size<=sizeof(float))
	 mtype=H5Tcopy(H5T_NATIVE_FLOAT);
      else if(size<=sizeof(double))
	 mtype=H5Tcopy(H5T_NATIVE_DOUBLE);
      else
	 mtype=H5Tcopy(H5T_NATIVE_LDOUBLE);
      break;
   case H5T_STRING:
      HL_SPEWDEBUG0("This is of type H5T_STRING");
      mtype=H5Tcopy(H5T_C_S1);
      H5Tset_size(mtype,size);
      strpad=H5Tget_strpad(type);
      H5Tset_strpad(mtype,strpad);
      if(H5Tequal(mtype,type)<0) {
	 H5Tclose(mtype);
	 mtype=H5Tcopy(H5T_FORTRAN_S1);
	 H5Tset_size(mtype,size);
	 H5Tset_strpad(mtype,strpad);
	 if(H5Tequal(mtype,type)<0)
	    mtype=-1;
      }
      break;
   case H5T_COMPOUND:
      HL_SPEWDEBUG0("This is of type H5T_COMPOUND");
      nmembs = H5Tget_nmembers(type);
      memb = calloc(nmembs, sizeof(hid_t));
      name = calloc(nmembs, sizeof(char*));
      ndims = calloc(nmembs, sizeof(int));
      dims141 = calloc(nmembs*4,sizeof(hsize_t));

      for(i=0;i<nmembs;i++) {
	 name[i]=NULL;
      }

      for(i=0,size=0;i<nmembs;i++) {
	 f_memb = H5Tget_member_type(type,i);
	 memb[i] = getFixedType(f_memb);

	 if(memb[i]<0) {
	    if(f_memb>=0) H5Tclose(f_memb);
	    goto done;
	 }

	 if (H5T_ARRAY==H5Tget_member_class(type,i)) {
	    hsize_t tmpdims[4];
	    H5Tget_array_dims(f_memb,tmpdims,NULL);
	    ndims[i] = H5Tget_array_ndims(f_memb);
	    if(ndims[i]<0 || ndims[i]>4) {
	       HL_ERROR0("Number of dims (the rank) not between 0-4");
	       if(f_memb>=0)  H5Tclose(f_memb);	       
	       goto done;
	    }
	    for(j=0,nelmts=1;j<ndims[i];j++) {
	       dims141[i*4+j] = tmpdims[j];
	       nelmts *= tmpdims[j];
	    }
	 }
	 else {
	    nelmts=1;
	 }	 

	 H5Tclose(f_memb);
	 
	 name[i] = H5Tget_member_name(type,i);
	 if(name[i]==NULL) {
	    goto done;
	 }
	 size = ALIGN(size, H5Tget_size(memb[i])) + nelmts*H5Tget_size(memb[i]);
      }
      
      mtype=H5Tcreate(H5T_COMPOUND,size);

      for(i=0,offset=0;i<nmembs;i++) {
	 for(j=0,nelmts=1;j<ndims[i];j++)
	    nelmts *= dims141[i*4+j];

	 if (H5T_ARRAY==H5Tget_member_class(type,i)) {
	    f_memb = H5Tget_member_type(type,i);

	    member_type=H5Tarray_create(memb[i],ndims[i],dims141+i*4,NULL);
	    H5Tinsert(mtype,name[i],offset,member_type);
	    H5Tclose(member_type);

	    /* offset=ALIGN(offset, H5Tget_size(memb[i])) +
	     * nelmts * H5Tget_size(memb[i]);
	     * What is more correct? This, or the following:?
	     * - Adam Dybbroe, 2001-07-12
	     */
	    offset=ALIGN(offset, H5Tget_size(memb[i])) + H5Tget_size(f_memb);
	    if(f_memb>=0) H5Tclose(f_memb);
	 }
	 else {
	    H5Tinsert(mtype,name[i],offset,memb[i]);
	    offset=ALIGN(offset, H5Tget_size(memb[i])) + nelmts*H5Tget_size(memb[i]);
	 }	    
      }

      break;
   case H5T_ENUM:
      HL_SPEWDEBUG0("This is of type H5T_ENUM");
      break;
   case H5T_REFERENCE:
      HL_SPEWDEBUG0("This is of type H5T_REFERENCE");
      break;
   case H5T_OPAQUE:
      HL_SPEWDEBUG0("This is of type H5T_OPAQUE");
      mtype=H5Tcopy(type);
      break;
   case H5T_BITFIELD:
      HL_SPEWDEBUG0("This is of type H5T_BITFIELD");
      mtype=H5Tcopy(type);
      H5Tset_offset(mtype,0);
      H5Tset_order(mtype,H5T_ORDER_LE);
      break;
   case H5T_TIME:
      HL_SPEWDEBUG0("This is of type H5T_TIME");
      break;
   case H5T_ARRAY:
      HL_SPEWDEBUG0("This is of type H5T_ARRAY");
      if((mtype=H5Tget_super(type))<0) {
	 HL_ERROR0("Failed getting the type id for the type that the array is made up of...");
	 goto done;
      }
      tmpt=mtype;
      mtype=getFixedType(mtype);
      if(H5Tclose(tmpt)<0) {
	 HL_ERROR0("Failed closing super type");
      }
      break;

   default:
      HL_INFO0("HRRM, I don't recognize this datatype");
      break;
   }

 done:
   if(memb && name && ndims && dims141) {
      for(i=0;i<nmembs;i++) {
	 if(memb[i]>=0)
	    H5Tclose(memb[i]);
	 if(name[i])
	    free(name[i]);
      }
      free(memb);
      free(name);
      free(ndims);
      if (dims141) free(dims141);
   }

   return mtype;
}

#else
/* VERSION 1.2.2 or less:	*/
hid_t getFixedType(hid_t type)
{
   size_t size;
   hid_t mtype=-1;
   H5T_str_t strpad;
   hid_t* memb=NULL;
   char** name=NULL;
   int nmembs=0,i,j,*ndims=NULL;
   size_t offset,nelmts;
   size_t *dims=NULL;
   hid_t f_memb=-1;

   HL_SPEWDEBUG0("ENTER getFixedType, version 1.2.2 or less");
   size=H5Tget_size(type);

   switch(H5Tget_class(type)) {
   case H5T_INTEGER:
      if(size<=sizeof(char))
	 mtype=H5Tcopy(H5T_NATIVE_SCHAR);
      else if(size<=sizeof(short))
	 mtype=H5Tcopy(H5T_NATIVE_SHORT);
      else if(size<=sizeof(int))
	 mtype=H5Tcopy(H5T_NATIVE_INT);
      else if(size<=sizeof(long))
	 mtype=H5Tcopy(H5T_NATIVE_LONG);
      else
	 mtype=H5Tcopy(H5T_NATIVE_LLONG);
      H5Tset_sign(mtype,H5Tget_sign(type));
      break;
   case H5T_FLOAT:
      if(size<=sizeof(float))
	 mtype=H5Tcopy(H5T_NATIVE_FLOAT);
      else if(size<=sizeof(double))
	 mtype=H5Tcopy(H5T_NATIVE_DOUBLE);
      else
	 mtype=H5Tcopy(H5T_NATIVE_LDOUBLE);
      break;
   case H5T_STRING:
      mtype=H5Tcopy(H5T_C_S1);
      H5Tset_size(mtype,size);
      strpad=H5Tget_strpad(type);
      H5Tset_strpad(mtype,strpad);
      if(H5Tequal(mtype,type)<0) {
	 H5Tclose(mtype);
	 mtype=H5Tcopy(H5T_FORTRAN_S1);
	 H5Tset_size(mtype,size);
	 H5Tset_strpad(mtype,strpad);
	 if(H5Tequal(mtype,type)<0)
	    mtype=-1;
      }
      break;
   case H5T_COMPOUND:
      nmembs = H5Tget_nmembers(type);
      memb = calloc(nmembs, sizeof(hid_t));
      name = calloc(nmembs, sizeof(char*));
      ndims = calloc(nmembs, sizeof(int));
      dims = calloc(nmembs*4,sizeof(size_t));

      for(i=0;i<nmembs;i++) {
	 name[i]=NULL;
      }

      for(i=0,size=0;i<nmembs;i++) {
	 f_memb = H5Tget_member_type(type,i);
	 memb[i] = getFixedType(f_memb);
	 H5Tclose(f_memb);
	 if(memb[i]<0)
	    goto done;

	 ndims[i] = H5Tget_member_dims(type,i,dims+i*4,NULL);
	 if(ndims[i]<0 || ndims[i]>4) {
	    HL_ERROR0("Number of dims (the rank) not between 0-4");
	    goto done;
	 }
	 for(j=0,nelmts=1;j<ndims[i];j++) {
	    nelmts *= dims[i*4+j];
	 }

	 /* Incompatible with version 1.4 and higher - Adam Dybbroe, 2001-05-16 */
	 name[i] = H5Tget_member_name(type,i);
	 if(name[i]==NULL) {
	    goto done;
	 }
	 size = ALIGN(size, H5Tget_size(memb[i])) + nelmts*H5Tget_size(memb[i]);
      }
      
      mtype=H5Tcreate(H5T_COMPOUND,size);
      for(i=0,offset=0;i<nmembs;i++) {
	 H5Tinsert_array(mtype,name[i],offset,ndims[i],dims+i*4,NULL,memb[i]);
	 /*
	   Incompatible with version 1.4 and higher. Adam Dybbroe, 2001-04-26
	 */
	 for(j=0,nelmts=1;j<ndims[i];j++)
	    nelmts *= dims[i*4+j];

	 offset=ALIGN(offset, H5Tget_size(memb[i])) +
	    nelmts*H5Tget_size(memb[i]);
      }
      break;
   case H5T_ENUM:
      break;
   case H5T_REFERENCE:
      break;
   case H5T_OPAQUE:
      mtype=H5Tcopy(type);
      break;
   case H5T_BITFIELD:
      mtype=H5Tcopy(type);
      H5Tset_offset(mtype,0);
      H5Tset_order(mtype,H5T_ORDER_LE);
      break;
   case H5T_TIME:
      break;
   default:
      HL_INFO0("HRRM, I don't recognize this datatype");
      break;
   }

 done:
   if(memb && name && ndims && dims) {
      for(i=0;i<nmembs;i++) {
	 if(memb[i]>=0)
	    H5Tclose(memb[i]);
	 if(name[i])
	    free(name[i]);
      }
      free(memb);
      free(name);
      free(ndims);
      free(dims);
   }

   return mtype;
}
#endif
#else
hid_t getFixedType(hid_t type)
{
   HL_CRITICAL0("No H5_VERS_MINOR defined!!!");
   return NULL;
}
#endif	 

/************************************************
 * translateCharToDatatype
 ***********************************************/
hid_t translateCharToDatatype(const char* dataType)
{
   hid_t retv=-1;
   HL_DEBUG0("ENTER: translateCharToDatatype");
   
   if(!dataType) {
      HL_ERROR0("No type name specified");
      return retv;
   }
   if(strcmp(dataType,"char")==0)
      retv=H5Tcopy(H5T_NATIVE_CHAR);
   else if(strcmp(dataType,"schar")==0)
      retv=H5Tcopy(H5T_NATIVE_SCHAR);
   else if(strcmp(dataType,"uchar")==0)
      retv=H5Tcopy(H5T_NATIVE_UCHAR);
   else if(strcmp(dataType,"short")==0)
      retv=H5Tcopy(H5T_NATIVE_SHORT);
   else if(strcmp(dataType,"ushort")==0)
      retv=H5Tcopy(H5T_NATIVE_USHORT);
   else if(strcmp(dataType,"int")==0)
      retv=H5Tcopy(H5T_NATIVE_INT);
   else if(strcmp(dataType,"uint")==0)
      retv=H5Tcopy(H5T_NATIVE_UINT);
   else if(strcmp(dataType,"long")==0)
      retv=H5Tcopy(H5T_NATIVE_LONG);
   else if(strcmp(dataType,"ulong")==0)
      retv=H5Tcopy(H5T_NATIVE_ULONG);
   else if(strcmp(dataType,"llong")==0)
      retv=H5Tcopy(H5T_NATIVE_LLONG);
   else if(strcmp(dataType,"ullong")==0)
      retv=H5Tcopy(H5T_NATIVE_ULLONG);
   else if(strcmp(dataType,"float")==0)
      retv=H5Tcopy(H5T_NATIVE_FLOAT);
   else if(strcmp(dataType,"double")==0)
      retv=H5Tcopy(H5T_NATIVE_DOUBLE);
   else if(strcmp(dataType,"ldouble")==0)
      retv=H5Tcopy(H5T_NATIVE_LDOUBLE);
   else if(strcmp(dataType,"hsize")==0)
      retv=H5Tcopy(H5T_NATIVE_HSIZE);
   else if(strcmp(dataType,"hssize")==0)
      retv=H5Tcopy(H5T_NATIVE_HSSIZE);
   else if(strcmp(dataType,"herr")==0)
      retv=H5Tcopy(H5T_NATIVE_HERR);
   else if(strcmp(dataType,"hbool")==0)
      retv=H5Tcopy(H5T_NATIVE_HBOOL);

   if(retv==-1) {
      HL_ERROR1("There is no type called %s",dataType);
   }
   return retv;
}

/************************************************
 * getTypeNameString
 ***********************************************/
char* getTypeNameString(hid_t type)
{
   char* retv=NULL;
   HL_DEBUG0("ENTER: getTypeNameString");
   
   switch(H5Tget_class(type)) {
   case H5T_INTEGER:
      if(H5Tequal(type,H5T_STD_I8BE))
	 retv=strdup("H5T_STD_I8BE");
      else if(H5Tequal(type, H5T_STD_I8LE))
	 retv=strdup("H5T_STD_I8LE");
      else if (H5Tequal(type, H5T_STD_I16BE))
	 retv=strdup("H5T_STD_I16BE");
      else if (H5Tequal(type, H5T_STD_I16LE))
	 retv=strdup("H5T_STD_I16LE");
      else if (H5Tequal(type, H5T_STD_I32BE))
	 retv=strdup("H5T_STD_I32BE");
      else if (H5Tequal(type, H5T_STD_I32LE))
	 retv=strdup("H5T_STD_I32LE");
      else if (H5Tequal(type, H5T_STD_I64BE))
	 retv=strdup("H5T_STD_I64BE");
      else if (H5Tequal(type, H5T_STD_I64LE))
	 retv=strdup("H5T_STD_I64LE");
      else if (H5Tequal(type, H5T_STD_U8BE))
	 retv=strdup("H5T_STD_U8BE");
      else if (H5Tequal(type, H5T_STD_U8LE))
	 retv=strdup("H5T_STD_U8LE");
      else if (H5Tequal(type, H5T_STD_U16BE))
	 retv=strdup("H5T_STD_U16BE");
      else if (H5Tequal(type, H5T_STD_U16LE))
	 retv=strdup("H5T_STD_U16LE");
      else if (H5Tequal(type, H5T_STD_U32BE))
	 retv=strdup("H5T_STD_U32BE");
      else if (H5Tequal(type, H5T_STD_U32LE))
	 retv=strdup("H5T_STD_U32LE");
      else if (H5Tequal(type, H5T_STD_U64BE))
	 retv=strdup("H5T_STD_U64BE");
      else if (H5Tequal(type, H5T_STD_U64LE))
	 retv=strdup("H5T_STD_U64LE");
      else if (H5Tequal(type, H5T_NATIVE_SCHAR)) 
	 retv=strdup("H5T_NATIVE_SCHAR");
      else if (H5Tequal(type, H5T_NATIVE_UCHAR))
	 retv=strdup("H5T_NATIVE_UCHAR");
      else if (H5Tequal(type, H5T_NATIVE_SHORT))
	 retv=strdup("H5T_NATIVE_SHORT");
      else if (H5Tequal(type, H5T_NATIVE_USHORT))
	 retv=strdup("H5T_NATIVE_USHORT");
      else if (H5Tequal(type, H5T_NATIVE_INT))
	 retv=strdup("H5T_NATIVE_INT");
      else if (H5Tequal(type, H5T_NATIVE_UINT))
	 retv=strdup("H5T_NATIVE_UINT");
      else if (H5Tequal(type, H5T_NATIVE_LONG))
	 retv=strdup("H5T_NATIVE_LONG");
      else if (H5Tequal(type, H5T_NATIVE_ULONG))
	 retv=strdup("H5T_NATIVE_ULONG");
      else if (H5Tequal(type, H5T_NATIVE_LLONG))
	 retv=strdup("H5T_NATIVE_LLONG");
      else if (H5Tequal(type, H5T_NATIVE_ULLONG))
	 retv=strdup("H5T_NATIVE_ULLONG");
      else {
	 HL_ERROR0("Undefined integer type");
      }
      break;
   case H5T_FLOAT:
      if (H5Tequal(type, H5T_IEEE_F32BE))
	 retv=strdup("H5T_IEEE_F32BE");
      else if (H5Tequal(type, H5T_IEEE_F32LE))
	 retv=strdup("H5T_IEEE_F32LE");
      else if (H5Tequal(type, H5T_IEEE_F64BE))
	 retv=strdup("H5T_IEEE_F64BE");
      else if (H5Tequal(type, H5T_IEEE_F64LE))
	 retv=strdup("H5T_IEEE_F64LE");
      else if (H5Tequal(type, H5T_NATIVE_FLOAT))
	 retv=strdup("H5T_NATIVE_FLOAT");
      else if (H5Tequal(type, H5T_NATIVE_DOUBLE))
	 retv=strdup("H5T_NATIVE_DOUBLE");
      else if (H5Tequal(type, H5T_NATIVE_LDOUBLE))
	 retv=strdup("H5T_NATIVE_LDOUBLE");
      else {
	 HL_ERROR0("Undefined float type");
      }
      break;
   case H5T_TIME:
      HL_INFO0("Unsupported type: H5T_TIME");
      break;
   case H5T_STRING:
      retv=strdup("H5T_STRING");
      break;
   case H5T_COMPOUND:
      retv=strdup("H5T_COMPOUND");
      break;
#ifdef H5_VERS_MINOR
      /*#if H5_VERS_MINOR>=4*/
#if H5_VERS_MAJOR>1 || (H5_VERS_MAJOR<2 && H5_VERS_MINOR>=4)
   case H5T_ARRAY:
      retv=strdup("H5T_ARRAY");
      break;
#endif
#endif
   default:
      HL_INFO0("Unsupported class");
      break;
   }

   return retv;
}

/************************************************
 * getFormatNameString
 ***********************************************/
char* getFormatNameString(hid_t type)
{
   char* retv=NULL;

   HL_SPEWDEBUG0("ENTER: getFormatNameString");
   
   if(H5Tget_class(type)==H5T_STRING) {
      retv=strdup("string");
      return retv;
   }

   if(H5Tequal(type,H5T_NATIVE_SCHAR))
      retv=strdup("schar");
   else if(H5Tequal(type,H5T_NATIVE_UCHAR))
      retv=strdup("uchar");
   else if(H5Tequal(type,H5T_NATIVE_CHAR))
      retv=strdup("char");
   else if(H5Tequal(type,H5T_NATIVE_SHORT))
      retv=strdup("short");
   else if(H5Tequal(type,H5T_NATIVE_USHORT))
      retv=strdup("ushort");
   else if(H5Tequal(type,H5T_NATIVE_INT))
      retv=strdup("int");
   else if(H5Tequal(type,H5T_NATIVE_UINT))
      retv=strdup("uint");
   else if(H5Tequal(type,H5T_NATIVE_LONG))
      retv=strdup("long");
   else if(H5Tequal(type,H5T_NATIVE_ULONG))
      retv=strdup("ulong");
   else if(H5Tequal(type,H5T_NATIVE_LLONG))
      retv=strdup("llong");
   else if(H5Tequal(type,H5T_NATIVE_ULLONG))
      retv=strdup("ullong");
   else if(H5Tequal(type,H5T_NATIVE_FLOAT))
      retv=strdup("float");
   else if(H5Tequal(type,H5T_NATIVE_DOUBLE))
      retv=strdup("double");
   else if(H5Tequal(type,H5T_NATIVE_LDOUBLE))
      retv=strdup("ldouble");
   else if(H5Tequal(type,H5T_NATIVE_HSIZE))
      retv=strdup("hsize");
   else if(H5Tequal(type,H5T_NATIVE_HSSIZE))
      retv=strdup("hssize");
   else if(H5Tequal(type,H5T_NATIVE_HERR))
      retv=strdup("herr");
   else if(H5Tequal(type,H5T_NATIVE_HBOOL))
      retv=strdup("hbool");
   else if(H5Tget_class(type)==H5T_COMPOUND)
      retv=strdup("compound");
#ifdef H5_VERS_MINOR
/*#if H5_VERS_MINOR>=4*/
#if H5_VERS_MAJOR>1 || (H5_VERS_MAJOR<2 && H5_VERS_MINOR>=4)
   else if(H5Tget_class(type)==H5T_ARRAY)
      retv=strdup("array");
#endif
#endif
   else {
      HL_INFO0("Not possible to translate from given type to string");

   }
   return retv;
}

/************************************************
 * getStringPadName
 ***********************************************/
char* getStringPadName(hid_t type)
{
   H5T_str_t strpad=H5Tget_strpad(type);
   char* retv=NULL;
   HL_DEBUG0("ENTER: getStringPadName");
   
   if(strpad==H5T_STR_NULLTERM)
      retv=strdup("H5T_STR_NULLTERM");
   else if(strpad==H5T_STR_NULLPAD)
      retv=strdup("H5T_STR_NULLPAD");
   else if(strpad==H5T_STR_SPACEPAD)
      retv=strdup("H5T_STR_SPACEPAD");
   else
      retv=strdup("ILLEGAL STRPAD");

   return retv;
}

/************************************************
 * getStringCsetName
 ***********************************************/
char* getStringCsetName(hid_t type)
{
   H5T_cset_t cset=H5Tget_cset(type);
   char* retv=NULL;
   HL_DEBUG0("ENTER: getStringCsetName");
   if(cset==H5T_CSET_ASCII)
      retv=strdup("H5T_CSET_ASCII");
   else
      retv=strdup("UNKNOWN CHARACTER SET");

   return retv;
}

/************************************************
 * getStringCtypeName
 ***********************************************/
char* getStringCtypeName(hid_t type)
{
   hid_t strtype=H5Tcopy(H5T_C_S1);
   char* retv=NULL;
   H5T_cset_t cset=H5Tget_cset(type);
   H5T_str_t strpad=H5Tget_strpad(type);
   size_t size=H5Tget_size(type);
   H5Tset_cset(strtype, cset);
   H5Tset_size(strtype, size);
   H5Tset_strpad(strtype, strpad);

   HL_DEBUG0("ENTER: getStringCtypeName");
   if(H5Tequal(type,strtype)) {
      retv=strdup("H5T_C_S1");
      H5Tclose(strtype);
   } else {
      H5Tclose(strtype);
      strtype=H5Tcopy(H5T_FORTRAN_S1);
      H5Tset_cset(strtype, cset);
      H5Tset_size(strtype, size);
      H5Tset_strpad(strtype, strpad);
      if(H5Tequal(type,strtype))
	 retv=strdup("H5T_FORTRAN_S1");
      else
	 retv=strdup("UNKNOWN CHARACTER TYPE");
      H5Tclose(strtype);
   }
   return retv;
}

/************************************************
 * newGroup
 ***********************************************/
NameListGroup_t* newGroup(NameListGroup_t* parentGroup,const char* name)
{
   NameListGroup_t* retgrp=NULL;
   HL_DEBUG0("ENTER: newGroup");
   
   if(parentGroup==NULL) {
      retgrp=(NameListGroup_t*)malloc(sizeof(NameListGroup_t));
      retgrp->name=strdup(name);

      retgrp->noGroups=0;
      retgrp->noallocGroups=0;
      retgrp->groups=NULL;

      retgrp->noDatasets=0;
      retgrp->noallocDatasets=0;
      retgrp->datasets=NULL;

      retgrp->noAttrs=0;
      retgrp->noallocAttrs=0;
      retgrp->attrs=NULL;

      retgrp->noGlobalTypes=0;
      retgrp->noallocGlobalTypes=0;
      retgrp->globalTypes=NULL;

      retgrp->noLocalTypes=0;
      retgrp->noallocLocalTypes=0;
      retgrp->localTypes=NULL;

      retgrp->rootGroup=retgrp; /*Since if parentGroup==NULL, then this is the root group
				 *be aware, that this is a bit hazardous, when deallocating
				 *memory*/
      retgrp->thisIsRootGrp=1;
   } else {
      int i;
      if(parentGroup->noallocGroups==parentGroup->noGroups ||
	 parentGroup->noallocGroups==0) {
	 int newsize=parentGroup->noallocGroups*2;
	 if(parentGroup->noallocGroups==0)
	    newsize = DEFAULT_NOOF_CHILDS;
	 parentGroup->groups=realloc(parentGroup->groups,newsize*sizeof(NameListGroup_t));
	 if(parentGroup->groups==NULL) {
	    HL_ERROR0("Failed to allocate memory for sub groups");
	    return NULL;
	 }
	 for(i=parentGroup->noallocGroups;i<newsize;i++) {
	    parentGroup->groups[i].name=NULL;

	    parentGroup->groups[i].noGroups=0;
	    parentGroup->groups[i].noallocGroups=0;
	    parentGroup->groups[i].groups=NULL;

	    parentGroup->groups[i].noDatasets=0;
	    parentGroup->groups[i].noallocDatasets=0;
	    parentGroup->groups[i].datasets=NULL;

	    parentGroup->groups[i].noAttrs=0;
	    parentGroup->groups[i].noallocAttrs=0;
	    parentGroup->groups[i].attrs=NULL;

	    parentGroup->groups[i].noGlobalTypes=0;
	    parentGroup->groups[i].noallocGlobalTypes=0;
	    parentGroup->groups[i].globalTypes=NULL;

	    parentGroup->groups[i].noLocalTypes=0;
	    parentGroup->groups[i].noallocLocalTypes=0;
	    parentGroup->groups[i].localTypes=NULL;

	    parentGroup->groups[i].rootGroup=parentGroup->rootGroup;

	    parentGroup->groups[i].thisIsRootGrp=0;
	 }
	 parentGroup->noallocGroups=newsize;

	 retgrp=&parentGroup->groups[parentGroup->noGroups];
	 retgrp->name=strdup(name);
	 parentGroup->noGroups++;
      } else if (parentGroup->noallocGroups<=parentGroup->noGroups) {
	 HL_ERROR0("Internal programming error in groups");
	 return NULL;
      } else {
	 /*No worries, should look like this*/
	 retgrp=&parentGroup->groups[parentGroup->noGroups];
	 retgrp->name=strdup(name);
	 parentGroup->noGroups++;
      }
   }

   return retgrp;
}

/************************************************
 * newDataset
 ***********************************************/
NameListDataset_t* newDataset(NameListGroup_t* parentGroup, const char* name)
{
   NameListDataset_t* retdset=NULL;
   int i;
   HL_DEBUG0("ENTER: newDataset");
   if(parentGroup==NULL) {
      HL_ERROR0("Programming error, trying to create a dataset on a nonexisting group");
      return NULL;
   }

   if(parentGroup->noallocDatasets == parentGroup->noDatasets ||
      parentGroup->noallocDatasets == 0) {
      int newsize = parentGroup->noallocDatasets*2;
      if(parentGroup->noallocDatasets==0)
	 newsize = DEFAULT_NOOF_CHILDS;
      parentGroup->datasets=realloc(parentGroup->datasets,newsize*sizeof(NameListDataset_t));
      if(parentGroup->datasets==NULL) {
	 HL_ERROR0("Failed to allocate memory for datasets in group");
	 return NULL;
      }
      for(i=parentGroup->noallocDatasets;i<newsize;i++) {
	 parentGroup->datasets[i].name=NULL;
	 parentGroup->datasets[i].ndims=0;
	 parentGroup->datasets[i].dims=NULL;
	 parentGroup->datasets[i].data=NULL;
	 parentGroup->datasets[i].datatype=NULL;
	 parentGroup->datasets[i].noAttrs=0;
	 parentGroup->datasets[i].noallocAttrs=0;
	 parentGroup->datasets[i].attrs=NULL;
	 parentGroup->datasets[i].rootGroup=parentGroup->rootGroup;
      }
      parentGroup->noallocDatasets=newsize;
      retdset=&parentGroup->datasets[parentGroup->noDatasets];
      retdset->name=strdup(name);
      parentGroup->noDatasets++;
   } else if(parentGroup->noallocDatasets<=parentGroup->noDatasets) {
      HL_ERROR0("Programming error in datasets");
      return NULL;
   } else {
      /*No worries, should look like this*/
      retdset=&parentGroup->datasets[parentGroup->noDatasets];
      retdset->name=strdup(name);
      parentGroup->noDatasets++;
   }

   return retdset;
}

/************************************************
 * newNameListType
 ***********************************************/
NameListType_t* newNameListType()
{
   NameListType_t* retv=(NameListType_t*)malloc(sizeof(NameListType_t));

   HL_DEBUG0("ENTER: newNameListType");
   if(!retv)
      return NULL;

   retv->referenceId=0;
   retv->typeSize=0;
   retv->format_type=NULL;
   retv->ieee_type=NULL;
   retv->isString=0;
   retv->strpad=NULL;
   retv->strcset=NULL;
   retv->strctype=NULL;
   retv->isCompound=0;
   retv->noAttrs=0;
   retv->noallocAttrs=0;
   retv->isNamed=0;
   retv->isOwnedByGroup=0;
   retv->name=NULL;
   retv->attrs=NULL;
   retv->objno[0]=retv->objno[1]=0;

   return retv;
}

/************************************************
 * newAttribute
 ***********************************************/
NameListAttribute_t* newAttribute(const char* name)
{
   NameListAttribute_t* attr=(NameListAttribute_t*)malloc(sizeof(NameListAttribute_t));
   HL_DEBUG0("ENTER: newAttribute");
   if(!attr)
      return NULL;

   attr->name=NULL;

   if(name)
      attr->name=strdup(name);

   attr->ndims=0;
   attr->dims=NULL;
   attr->data=NULL;
   attr->attrtype=NULL;

   return attr;
}

/************************************************
 * newCompoundAttribute
 ***********************************************/
CompoundAttributeDef_t* newCompoundAttribute(const char* name)
{
   CompoundAttributeDef_t* retv = (CompoundAttributeDef_t*)malloc(sizeof(CompoundAttributeDef_t));
   HL_DEBUG0("ENTER: newCompoundAttribute");
   
   if(!retv) {
      HL_ERROR0("Failed to allocate memory for CompoundAttributeDef_t");
      return NULL;
   }
   retv->name=NULL;
   retv->attrSize = 0;
   retv->format_type=NULL;
   retv->ieee_type=NULL;
   retv->ndims=0;
   retv->dims[0]=retv->dims[1]=retv->dims[2]=retv->dims[3]=0;
   retv->isString=0;
   retv->strpad=NULL;
   retv->strcset=NULL;
   retv->strctype=NULL;
   retv->offset=0;

   if(name)
      retv->name = strdup(name);

   return retv;
}

/************************************************
 * createCompoundFromType
 ***********************************************/
CompoundAttributeDef_t* createCompoundFromType(NameListType_t* inType, char* name)
{
   CompoundAttributeDef_t* retv=NULL;
   HL_DEBUG0("ENTER: createCompoundFromType");
   if(!inType) {
      HL_ERROR0("inType to createCompoundFromType is NULL, will create blank");
      retv=newCompoundAttribute(name);
      return retv;
   }
   /*printf("createCompoundFromType: name = %s\n", name);*/
   
   retv = newCompoundAttribute(name);

   /*   printf("inType->typeSize = %d\n", inType->typeSize);
   printf("inType->isString = %d\n", inType->isString);
   */
   retv->attrSize = inType->typeSize;
   retv->isString = inType->isString;
   retv->offset = 0;

   if(inType->format_type)
      retv->format_type = strdup(inType->format_type);
   if(inType->ieee_type)
      retv->ieee_type = strdup(inType->ieee_type);
   if(inType->strpad)
      retv->strpad = strdup(inType->strpad);
   if(inType->strcset)
      retv->strcset = strdup(inType->strcset);
   if(inType->strctype)
      retv->strctype = strdup(inType->strctype);

   /*
   printf("inType->format_type, retv->format_type: %s %s\n", inType->format_type,retv->format_type);
   printf("inType->ieee_type, retv->ieee_type: %s %s\n", inType->ieee_type,retv->ieee_type);
   printf("inType->strpad, retv->strpad: %s %s\n", inType->strpad,retv->strpad);
   printf("inType->strcset, retv->strcset: %s %s\n", inType->strcset,retv->strcset);
   printf("inType->strctype, retv->strctype: %s %s\n", inType->strctype,retv->strctype);
   printf("createCompoundFromType: retv->name = %s\n", retv->name);
   */
   
   return retv;
}

/************************************************
 * addCompoundAttributeToType
 ***********************************************/
herr_t addCompoundAttributeToType(NameListType_t* newType,CompoundAttributeDef_t* compoundAttr)
{
   int i;
   HL_DEBUG0("ENTER: addCompoundAttributeToType");
   if(!newType) {
      HL_ERROR0("Trying to add a compound attribute to a non existing type");
      return -1;
   }

   if(newType->noallocAttrs == newType->noAttrs ||
      newType->noallocAttrs == 0) {
      int newsize = newType->noallocAttrs*2;
      if(newType->noallocAttrs==0)
	 newsize=DEFAULT_NOOF_CHILDS;
      newType->attrs=realloc(newType->attrs,newsize*sizeof(CompoundAttributeDef_t*));
      if(newType->attrs==NULL) {
	 HL_ERROR0("Failed to allocate memory for compound attrs in type");
	 return -1;
      }
      for(i=newType->noallocAttrs;i<newsize;i++) {
	 newType->attrs[i]=NULL;
      }
      newType->noallocAttrs=newsize;
      newType->attrs[newType->noAttrs]=compoundAttr;
      newType->noAttrs++;
      return 0;
   } else if(newType->noallocAttrs<=newType->noAttrs) {
      HL_ERROR0("Programming error in addCompoundAttributeToType");
      return -1;
   } else {
      newType->attrs[newType->noAttrs]=compoundAttr;
      newType->noAttrs++;
   }

   return 0;
}

/************************************************
 * addAttributeToGroup
 ***********************************************/
herr_t addAttributeToGroup(NameListGroup_t* group, NameListAttribute_t* attr)
{
   int i;
   HL_DEBUG0("ENTER: addAttributeToGroup");
   if(!group) {
      HL_ERROR0("Trying to add an attribute to a non existing group");
      return -1;
   }

   if(group->noallocAttrs==group->noAttrs ||
      group->noallocAttrs==0) {
      int newsize=group->noallocAttrs*2;
      if(group->noallocAttrs==0)
	 newsize=DEFAULT_NOOF_CHILDS;
      group->attrs=realloc(group->attrs,newsize*sizeof(NameListAttribute_t*));
      if(group->attrs==NULL) {
	 HL_ERROR0("Failed to allocate memory for attributes in group");
	 return -1;
      }
      for(i=group->noallocAttrs;i<newsize;i++) {
	 group->attrs[i]=NULL;
      }
      group->noallocAttrs=newsize;
      group->attrs[group->noAttrs]=attr;
      group->noAttrs++;
      return 0;
   } else if(group->noallocAttrs<=group->noAttrs) {
      HL_ERROR0("Programming error in add attribute to groups");
      return -1;
   } else {
      group->attrs[group->noAttrs]=attr;
      group->noAttrs++;
   }

   return 0;
}


/************************************************
 * addAttributeToDataset
 ***********************************************/
herr_t addAttributeToDataset(NameListDataset_t* dset, NameListAttribute_t* attr)
{
   int i;
   HL_DEBUG0("ENTER: addAttributeToDataset");
   if(!dset) {
      HL_ERROR0("Trying to add an attribute to a non existing group");
      return -1;
   }

   if(dset->noallocAttrs==dset->noAttrs ||
      dset->noallocAttrs==0) {
      int newsize=dset->noallocAttrs*2;
      if(dset->noallocAttrs==0)
	 newsize=DEFAULT_NOOF_CHILDS;
      dset->attrs=realloc(dset->attrs,newsize*sizeof(NameListAttribute_t*));
      if(dset->attrs==NULL) {
	 HL_ERROR0("Failed to allocate memory for attributes in group");
	 return -1;
      }
      for(i=dset->noallocAttrs;i<newsize;i++) {
	 dset->attrs[i]=NULL;
      }
      dset->noallocAttrs=newsize;
      dset->attrs[dset->noAttrs]=attr;
      dset->noAttrs++;
      return 0;
   } else if(dset->noallocAttrs<=dset->noAttrs) {
      HL_ERROR0("Programming error in add attribute to dataset");
      return -1;
   } else {
      dset->attrs[dset->noAttrs]=attr;
      dset->noAttrs++;
   }

   return 0;   
}

/************************************************
 * freeCompoundAttribute
 ***********************************************/
void freeCompoundAttribute(CompoundAttributeDef_t* attr)
{
   HL_DEBUG0("ENTER: freeCompoundAttribute");
   
   if(!attr) {
      HL_DEBUG0("Trying to deallocate NULL attribute");
      return;
   }

   if(attr->name)
      free(attr->name);
   if(attr->format_type)
      free(attr->format_type);
   if(attr->ieee_type)
      free(attr->ieee_type);
   if(attr->strpad)
      free(attr->strpad);
   if(attr->strcset)
      free(attr->strcset);
   if(attr->strctype)
      free(attr->strctype);

   free(attr);
}

/************************************************
 * freeInternalAttribute
 ***********************************************/
void freeInternalAttribute(NameListAttribute_t* attr)
{
   HL_DEBUG0("ENTER: freeInternalAttribute");
   
   if(!attr)
      return;
   if(attr->name) {
      free(attr->name);
   }
   if(attr->dims) {
      free(attr->dims);
   }
   if(attr->data) {
      free(attr->data);
   }
   if(attr->attrtype) {
      if(attr->attrtype->isOwnedByGroup<=0) {
	 freeNameListType(attr->attrtype);
      } else {
	 /*This datatype is owned by a group, ignore deleting it.*/
      }
   }
}


/************************************************
 * freeAttribute
 ***********************************************/
void freeAttribute(NameListAttribute_t* attr)
{
   HL_DEBUG0("ENTER: freeAttribute");
   freeInternalAttribute(attr);
   free(attr);
}

/************************************************
 * freeNameListType
 ***********************************************/
void freeNameListType(NameListType_t* type)
{
   HL_DEBUG0("ENTER: freeNameListType");
   if(!type)
      return;
   if(type->format_type)
      free(type->format_type);
   if(type->ieee_type)
      free(type->ieee_type);
   if(type->strpad)
      free(type->strpad);
   if(type->strcset)
      free(type->strcset);
   if(type->strctype)
      free(type->strctype);
   if(type->name)
      free(type->name);

   if(type->attrs) {
      int i;
      for(i=0;i<type->noAttrs;i++) {
	 freeCompoundAttribute(type->attrs[i]);
      }
      free(type->attrs);
   }

   free(type);
}

/************************************************
 * freeInternalDataset
 ***********************************************/
void freeInternalDataset(NameListDataset_t* dset)
{
   int i;
   HL_DEBUG0("ENTER: freeInternalDataset");
   if(dset->dims)
      free(dset->dims);
   if(dset->data)
      free(dset->data);
   if(dset->attrs) {
      for(i=0;i<dset->noAttrs;i++) {
	 freeAttribute(dset->attrs[i]);
      }
      free(dset->attrs);
   }

   if(dset->datatype) {
      if(dset->datatype->isOwnedByGroup<=0) {
	 freeNameListType(dset->datatype);
      } else {
	 /*This datatype is owned by a group, ignore deleting it.*/
      }
   }

   if(dset->name)
      free(dset->name);

}

/************************************************
 * freeDataset
 ***********************************************/
void freeDataset(NameListDataset_t* dset)
{
   HL_DEBUG0("ENTER: freeDataset");
   freeInternalDataset(dset);
   free(dset);
}

/************************************************
 * addTypeToLocalGroup
 ***********************************************/
herr_t addTypeToLocalGroup(NameListGroup_t* group, NameListType_t* type)
{
   int i;
   HL_DEBUG0("ENTER: addTypeToLocalGroup");
   if(!group) {
      HL_ERROR0("Trying to add a type to a non existing group");
      return -1;
   }

   if(group->noallocLocalTypes==group->noLocalTypes ||
      group->noallocLocalTypes==0) {
      int newsize=group->noallocLocalTypes*2;
      if(group->noallocLocalTypes==0)
	 newsize=DEFAULT_NOOF_CHILDS;
      group->localTypes=realloc(group->localTypes,newsize*sizeof(NameListType_t*));
      if(group->localTypes==NULL) {
	 HL_ERROR0("Failed to allocate memory for types in group");
	 return -1;
      }
      for(i=group->noallocLocalTypes;i<newsize;i++) {
	 group->localTypes[i]=NULL;
      }
      group->noallocLocalTypes=newsize;
      type->isOwnedByGroup++;
      group->localTypes[group->noLocalTypes]=type;
      group->noLocalTypes++;
      return 0;
   } else if(group->noallocLocalTypes<=group->noLocalTypes) {
      HL_ERROR0("Programming error in addTypeToLocalGroup");
      return -1;
   } else {
      type->isOwnedByGroup++;
      group->localTypes[group->noLocalTypes]=type;
      group->noLocalTypes++;
   }

   return 0;
}

/************************************************
 * addTypeToGlobalGroup
 ***********************************************/
herr_t addTypeToGlobalGroup(NameListGroup_t* group, NameListType_t* type)
{
   int i;
   HL_DEBUG0("ENTER: addTypeToGlobalGroup");
   if(!group) {
      HL_ERROR0("Trying to add a type to a non existing group");
      return -1;
   }

   if(group->noallocGlobalTypes==group->noGlobalTypes ||
      group->noallocGlobalTypes==0) {
      int newsize=group->noallocGlobalTypes*2;
      if(group->noallocGlobalTypes==0)
	 newsize=DEFAULT_NOOF_CHILDS;
      group->globalTypes=realloc(group->globalTypes,newsize*sizeof(NameListType_t*));
      if(group->globalTypes==NULL) {
	 HL_ERROR0("Failed to allocate memory for types in group");
	 return -1;
      }
      for(i=group->noallocGlobalTypes;i<newsize;i++) {
	 group->globalTypes[i]=NULL;
      }
      group->noallocGlobalTypes=newsize;
      type->isOwnedByGroup++;
      group->globalTypes[group->noGlobalTypes]=type;
      group->noGlobalTypes++;
      return 0;
   } else if(group->noallocGlobalTypes<=group->noGlobalTypes) {
      HL_ERROR0("Programming error in addTypeToGlobalGroup");
      return -1;
   } else {
      type->isOwnedByGroup++;
      group->globalTypes[group->noGlobalTypes]=type;
      group->noGlobalTypes++;
   }

   return 0;
}

/************************************************
 * doesTypeExistInGlobalGroup
 ***********************************************/
int doesTypeExistInGlobalGroup(NameListGroup_t* grp,unsigned long* objno)
{
   int i;
   HL_DEBUG0("ENTER: doesTypeExistInGlobalGroup");
   for(i=0;i<grp->noGlobalTypes;i++) {
      if(grp->globalTypes[i]->objno[0]==objno[0] &&
	 grp->globalTypes[i]->objno[1]==objno[1])
	 return i;
   }
   return -1;
}

/************************************************
 * doesTypeExistInLocalGroup
 ***********************************************/
int doesTypeExistInLocalGroup(NameListGroup_t* grp,unsigned long* objno)
{
   int i;
   HL_DEBUG0("ENTER: doesTypeExistInLocalGroup");
   for(i=0;i<grp->noLocalTypes;i++) {
      if(grp->localTypes[i]->objno[0]==objno[0] &&
	 grp->localTypes[i]->objno[1]==objno[1])
	 return i;
   }
   return -1;
}

/************************************************
 * removeTypeFromLocalGroup
 ***********************************************/
NameListType_t* removeTypeFromLocalGroup(NameListGroup_t* group,unsigned long* objno)
{
   int found=-1;
   int i;
   NameListType_t* retType=NULL;

   HL_DEBUG0("ENTER: removeTypeFromLocalGroup");
   found = doesTypeExistInLocalGroup(group,objno);
   if(found<0) {
      HL_ERROR0("Failed to locate the type in the local group defs");
      return NULL;
   }

   retType = group->localTypes[found];
   for(i=found+1;i<group->noLocalTypes;i++) {
      group->localTypes[i-1]=group->localTypes[i];
   }
   group->localTypes[group->noLocalTypes-1]=NULL;
   group->noLocalTypes--;

   retType->isOwnedByGroup--;

   return retType;
}

/************************************************
 * removeTypeFromGlobalGroup
 ***********************************************/
NameListType_t* removeTypeFromGlobalGroup(NameListGroup_t* group,unsigned long* objno)
{
   int found=-1;
   int i;
   NameListType_t* retType=NULL;

   HL_DEBUG0("ENTER: removeTypeFromGlobalGroup");
   found = doesTypeExistInGlobalGroup(group,objno);
   if(found<0) {
      HL_ERROR0("Failed to locate the type in the local group defs");
      return NULL;
   }

   retType = group->globalTypes[found];
   for(i=found+1;i<group->noGlobalTypes;i++) {
      group->globalTypes[i-1]=group->globalTypes[i];
   }

   group->globalTypes[group->noGlobalTypes-1]=NULL;
   group->noGlobalTypes--;

   retType->isOwnedByGroup--;

   return retType;
}

/**********************************************************
 *Function:	whatSizeIsHdfFormat
 **********************************************************/
int whatSizeIsHdfFormat(const char* format)
{
   hid_t tmpType;
   int size=-1;
   HL_DEBUG0("ENTER: whatSizeIsHdfFormat");
   tmpType=translateCharToDatatype(format);
   if(tmpType<0) {
      HL_ERROR1("There is no type called %s",format);
      return -1;
   }
   size=H5Tget_size(tmpType);
   H5Tclose(tmpType);
   return size;
}

/**********************************************************
 *Function:	isFormatSupported
 **********************************************************/
int isFormatSupported(const char* format)
{
   hid_t tmpHid=-1;
   int retv=1;
   HL_DEBUG0("ENTER: isFormatSupported");
   if((tmpHid=translateCharToDatatype(format))<0) {
      retv=0;
   }
   if(tmpHid>=0)
      H5Tclose(tmpHid);
   return retv;
}

/**********************************************************
 *Function:	newHL_Compression
 **********************************************************/
HL_Compression* newHL_Compression(HL_CompressionType aType)
{
   HL_Compression* retv = NULL;
   HL_DEBUG0("ENTER: newHL_Compression");
   if(!(retv=(HL_Compression*)malloc(sizeof(HL_Compression)))) {
      HL_ERROR0("Failed to allocate memory for HL_Compression");
      return NULL;
   }
   initHL_Compression(retv,aType);
   return retv;
}

/**********************************************************
 *Function:	dupHL_Compression
 **********************************************************/
HL_Compression* dupHL_Compression(HL_Compression* inv)
{
   HL_Compression* retv = NULL;
   HL_DEBUG0("ENTER: dupHL_Compression");
   if(!inv) {
      return NULL;
   }
   if(!(retv=(HL_Compression*)malloc(sizeof(HL_Compression)))) {
      HL_ERROR0("Failed to allocate memory for HL_Compression");
      return NULL;
   }
   retv->type = inv->type;
   retv->level = inv->level;
   retv->szlib_mask = inv->szlib_mask;
   retv->szlib_px_per_block = inv->szlib_px_per_block;
   return retv;
}

/**********************************************************
 *Function:	initHL_Compression
 **********************************************************/
void initHL_Compression(HL_Compression* inv,HL_CompressionType aType)
{
   HL_DEBUG0("ENTER: initHL_Compression");
   if(!inv) {
      HL_ERROR0("Trying to initialize NULL");
      return;
   }
   inv->type=aType;
   inv->level=6;
   inv->szlib_mask=H5_SZIP_ALLOW_K13_OPTION_MASK|H5_SZIP_EC_OPTION_MASK;
   inv->szlib_px_per_block=16;
}

/**********************************************************
 *Function:	freeHL_Compression
 **********************************************************/
void freeHL_Compression(HL_Compression* inv)
{
   HL_SPEWDEBUG0("ENTER: freeHL_Compression");
   if(!inv)
      return;
   free(inv);
}
