/* --------------------------------------------------------------------------
$Id: read_vhlhdf.c,v 1.9 2006/01/24 13:15:27 dmichels Exp $
Program:        

Description:    Used for reading an HDF5 file and building a node tree
		with the HighLevel protocol.

Author(s):      Anders Henja

Copyright:	Swedish Meteorological and Hydrological Institute, 2000

History:
$Log: read_vhlhdf.c,v $
Revision 1.9  2006/01/24 13:15:27  dmichels
Modified call to H5Dread in function fillDatasetNode. For reasons unknown,
our function became much slower in more recent versions of HDF5. It is now
RADICALLY faster, adding virtually no overhead to the basic HDF5 library.

Revision 1.8  2004/09/20 11:03:33  ahenja
Blehe..... removed #ifdef KALLE...

Revision 1.7  2004/09/20 09:58:16  ahenja
Fixed strange bug!?!?!?!
:wq!

Revision 1.6  2004/07/01 09:28:16  ahenja
Updated with support for new compression and new loghandling

Revision 1.5  2004/03/18 11:53:49  ahenja
Modified with support for szip, also changed so that it is possible
to specify HL_FileCreationProperty when creating a file. Also changed
how the compression handling is working. Added a nicer way to
perform debugging and error handling. Now it is possible to define
an own error handler and hook it into the system, if this is done
all error messages will be sent to the given function.

Revision 1.4  2003/01/16 09:36:16  ahenja
Added some missing code in buildTypeDescriptionFromTypeHid if
H5_VERS_MINOR not is defined, hopefully this should never
happen.

Revision 1.3  2002/12/19 13:29:14  ahenja
Added some typecastings when mallocing memory.

Revision 1.2  2002/08/06 10:29:17  ahenja
Corrected bug in write_hlhdf.c function addAttributeToCompoundType, an illegal
cast was performed for HDF5 Versions >= 1.4.x.

Revision 1.1.1.1  2001/12/14 12:02:40  ahenja
First CVS version of hlhdf

Revision 1.16  2001/12/07 13:20:32  ahenja
Updated with possibilities to read and write references.

Revision 1.15  2001/11/13 15:18:44  adybbroe
fixed for bug in createCompoundObject.

Revision 1.14  2001/11/08 12:25:02  ahenja
Slightly updated..

Revision 1.13  2001/11/08 08:29:24  adybbroe
Forced checkin since ahenja needs to perform updates.

Revision 1.12  2001/04/26 14:16:49  adybbroe
Solved for version 1.4 incompatibility.

Revision 1.11  2001/03/13 19:57:33  adybbroe
*** empty log message ***

Revision 1.10  2001/02/08 12:04:02  ahenja
Updated the compound object handling (types).

 * Revision 1.9  2001/01/18  12:40:24  ahenja
 * *** empty log message ***
 *
 * Revision 1.8  2000/03/22  08:38:19  ahenja
 * Updated with spell checking and header information.
 *
-----------------------------------------------------------------------------*/
#include "vhlhdf.h"
#include "read_vhlhdf.h"
#include "hlhdf_debug.h"
#include <time.h>

/* Struct used internally for beeing able to do a reverse
 * name lookup when reading a file with references.
 * We dont want to have the data in the object referenced
 * to, but instead we want to have the name, thats why this
 * has to be used.
 */
typedef struct ReferenceLookup {
   hid_t file_id;
   hobj_ref_t* ref;
   char tmp_name[512];
   char found_name[512];
   int reffound;
   struct ReferenceLookup* next;
} ReferenceLookup;

/* Forward declarations */
static int checkIfReferenceMatch(hid_t loc_id, char* path, hobj_ref_t* ref);
static herr_t refAttributeLocationIterator(hid_t loc_id, const char* name, void* op_data);
static herr_t refGroupLocationIterator(hid_t gid, const char* name, void* op_data);

/* ---------------------------------------
 * GROUP_ATTRIBUTE_ITERATOR
 * --------------------------------------- */
static herr_t groupAttributeIterator(hid_t loc_id, const char* name, void* op_data)
{
   HL_NodeList* nodeList=(HL_NodeList*)op_data;
   HL_Node* node;
   char tmp1[512],tmp2[512];
   hid_t attrid=-1,attrtype=-1;
   
   HL_SPEWDEBUG0("ENTER: groupAttributeIterator");
   
   attrid=H5Aopen_name(loc_id,name);   /* Open the attribute to check what class it is */
   attrtype=H5Aget_type(attrid);   /* Need this to be able to check class */
   strcpy(tmp1,nodeList->tmp_name);
   sprintf(tmp2,"%s/%s",nodeList->tmp_name,name);
   strcpy(nodeList->tmp_name,tmp2);
   if(H5Tget_class(attrtype)==H5T_REFERENCE) {
      /* Special case for attributes that really are references */
      addNode(nodeList, (node = newHL_Reference(nodeList->tmp_name)));
   } else {
      addNode(nodeList, (node = newHL_Attribute(nodeList->tmp_name)));
   }
   strcpy(nodeList->tmp_name,tmp1);
   H5Tclose(attrtype);
   H5Aclose(attrid);
   return 0;
}

/* ---------------------------------------
 * DATASET_ATTRIBUTE_ITERATOR (IS THIS REALLY NEEDED, 
 * IT IS THE SAME CODE AS FOR GROUP_ATTRIBUTE_ITERATOR)
 * --------------------------------------- */
static herr_t datasetAttributeIterator(hid_t loc_id, const char* name, void* op_data)
{
   HL_NodeList* nodeList=(HL_NodeList*)op_data;
   HL_Node* node;
   char tmp1[512],tmp2[512];
   hid_t attrid=-1,attrtype=-1;

   HL_SPEWDEBUG0("ENTER: datasetAttributeIterator");
   
   attrid=H5Aopen_name(loc_id,name);   /* Open the attribute to check what class it is */
   attrtype=H5Aget_type(attrid);   /* Need this to be able to check class */
   strcpy(tmp1,nodeList->tmp_name);
   sprintf(tmp2,"%s/%s",nodeList->tmp_name,name);
   strcpy(nodeList->tmp_name,tmp2);
   if(H5Tget_class(attrtype)==H5T_REFERENCE) {
      /* Special case for attributes that really are references */
      addNode(nodeList, (node = newHL_Reference(nodeList->tmp_name)));
   } else {
      addNode(nodeList, (node = newHL_Attribute(nodeList->tmp_name)));
   }
   strcpy(nodeList->tmp_name,tmp1);
   H5Tclose(attrtype);
   H5Aclose(attrid);
   return 0;   
}

/* ---------------------------------------
 * GROUP_ITERATOR
 * --------------------------------------- */
static herr_t groupIterator(hid_t gid, const char* name, void* op_data)
{
   hid_t obj;
   H5G_stat_t statbuf;
   HL_NodeList* nodeList=(HL_NodeList*)op_data;
   HL_Node* node;
   char tmp1[512],tmp2[512];

   HL_SPEWDEBUG0("ENTER: groupIterator");
   
   H5Gget_objinfo(gid,name,0,&statbuf);

   strcpy(tmp1,nodeList->tmp_name);

   switch(statbuf.type) {
   case H5G_GROUP:
      if((obj=H5Gopen(gid,name))>=0) {
	 sprintf(tmp2,"%s/%s",nodeList->tmp_name,name);
	 strcpy(nodeList->tmp_name,tmp2);
	 addNode(nodeList, (node = newHL_Group(nodeList->tmp_name)));
	 H5Aiterate(obj,NULL,groupAttributeIterator, nodeList);
	 H5Giterate(obj,".", NULL, groupIterator, nodeList);
	 strcpy(nodeList->tmp_name,tmp1);
	 H5Gclose(obj);
      }
      break;
   case H5G_DATASET:
      if((obj=H5Dopen(gid,name))>=0) {
	 sprintf(tmp2,"%s/%s",nodeList->tmp_name,name);
	 strcpy(nodeList->tmp_name,tmp2);
	 addNode(nodeList, (node=newHL_Dataset(nodeList->tmp_name)));
	 H5Aiterate(obj,NULL,datasetAttributeIterator,nodeList);
	 strcpy(nodeList->tmp_name,tmp1);
	 H5Dclose(obj);
      }
      break;
   case H5G_TYPE:
      sprintf(tmp2,"%s/%s",nodeList->tmp_name,name);
      strcpy(nodeList->tmp_name,tmp2);
      addNode(nodeList,(node=newHL_Datatype(nodeList->tmp_name)));
      strcpy(nodeList->tmp_name,tmp1);
      break;
   default:
      HL_ERROR1("Undefined type for %s",name);
      HL_ERROR1("Name: %s",tmp1);
      HL_ERROR1("Type id %d",statbuf.type);
      break;
   }
   return 0;
}

/* ---------------------------------------
 * READ_HL_NODE_LIST_FROM
 * --------------------------------------- */
HL_NodeList* readHL_NodeListFrom(const char* filename, const char* fromPath)
{
   hid_t file_id=-1,gid=-1;
   HL_NodeList* retv=NULL;

   HL_DEBUG0("ENTER: readHL_NodeListFrom");
   
   if((file_id=openHlHdfFile(filename,"r"))<0) {
      HL_ERROR1("Failed to open file %s",filename);
      return NULL;
   }

   if(!(retv=newHL_NodeList())) {
      HL_ERROR0("Could not allocate NodeList\n");
      goto fail;
   }
   strcpy(retv->filename,filename);
   if((gid=H5Gopen(file_id,fromPath))<0) {
      HL_ERROR0("Failed to open root group");
      goto fail;
   }
   H5Aiterate(gid,NULL,groupAttributeIterator,retv);
   H5Gclose(gid);
   gid=-1;

   if(H5Giterate(file_id,fromPath,NULL,groupIterator,retv)<0) {
      HL_ERROR0("Could not iterate over groups");
      goto fail;
   }

   markAllNodes(retv,NMARK_ORIGINAL);
   
   if(file_id>=0) H5Fclose(file_id);
   if(gid>=0) H5Gclose(gid);
   return retv;
fail:
   if(file_id>=0) H5Fclose(file_id);
   if(gid>=0) H5Gclose(gid);
   freeHL_NodeList(retv);
   return NULL;   
}

/* ---------------------------------------
 * READ_HL_NODE_LIST
 * --------------------------------------- */
HL_NodeList* readHL_NodeList(const char* filename)
{
   HL_DEBUG0("ENTER: readHL_NodeList");
   return readHL_NodeListFrom(filename,".");
}

/* ---------------------------------------
 * SELECT_NODE
 * --------------------------------------- */
int selectNode(HL_NodeList* nodelist, const char* name)
{
   int i;

   HL_DEBUG0("ENTER: selectNode");
   if(!name) {
      HL_ERROR0("Can not select any node when name is NULL");
      return 0;
   }

   for(i=0;i<nodelist->nNodes;i++) {
      if(strcmp(nodelist->nodes[i]->name,name)==0) {
	 nodelist->nodes[i]->mark = NMARK_SELECT;
	 return 1;
      }
   }

   HL_ERROR1("Could not find any node called '%s'",name);
   return 0;
}

/* ---------------------------------------
 * SELECT_ALL_NODES
 * --------------------------------------- */
int selectAllNodes(HL_NodeList* nodelist)
{
   int i;
   HL_DEBUG0("ENTER: selectAllNodes");
   for(i=0;i<nodelist->nNodes;i++) {
      nodelist->nodes[i]->mark = NMARK_SELECT;
   }

   return 1;
}

/* ---------------------------------------
 * SELECT_METADATA_NODES
 * VOLATILE: Do not attempt to access dataset arrays after calling this.
 * --------------------------------------- */
int selectMetadataNodes(HL_NodeList* nodelist)
{
   int i;
   HL_DEBUG0("ENTER: selectMetadataNodes");
   for(i=0;i<nodelist->nNodes;i++) {
      if ((nodelist->nodes[i]->type != DATASET_ID) && 
	  (nodelist->nodes[i]->dataType != HL_ARRAY)) {
	 nodelist->nodes[i]->mark = NMARK_SELECT;
      }
   }

   return 1;
}

/* ---------------------------------------
 * DE-SELECT_NODE
 * --------------------------------------- */
int deselectNode(HL_NodeList* nodelist, const char* name)
{
   int i;

   HL_DEBUG0("ENTER: deselectNode");
   if(!name) {
      HL_ERROR0("Can not deselect any node when name is NULL");
      return 0;
   }

   for(i=0;i<nodelist->nNodes;i++) {
      if(strcmp(nodelist->nodes[i]->name,name)==0) {
	 nodelist->nodes[i]->mark = NMARK_ORIGINAL;
	 return 1;
      }
   }

   HL_ERROR1("Could not find any node called '%s'",name);
   return 0;
}

/* ---------------------------------------
 * FETCH_MARKED_NODES
 * --------------------------------------- */
int fetchMarkedNodes(HL_NodeList* nodelist)
{
   int i;
   hid_t file_id=-1;
   hid_t gid=-1;
   HL_DEBUG0("ENTER: fetchMarkedNodes");
   if(!(file_id=openHlHdfFile(nodelist->filename,"r"))<0) {
      HL_ERROR1("Could not open file '%s' when fetching data",nodelist->filename);
      return 0;
   }

   if((gid=H5Gopen(file_id,"."))<0) {
      HL_ERROR0("Could not open root group\n");
      goto fail;
   }

   for(i=0;i<nodelist->nNodes;i++) {
      if(nodelist->nodes[i]->mark == NMARK_SELECT) {
	 if(!fillNodeWithData(file_id, nodelist->nodes[i])) {
	    HL_ERROR1("Error occured when trying to fill node '%s'",nodelist->nodes[i]->name);
	    goto fail;
	 }
      }
   }

   if(file_id>=0) H5Fclose(file_id);
   if(gid>=0) H5Gclose(gid);
   return 1;
 fail:
   if(file_id>=0) H5Fclose(file_id);
   if(gid>=0) H5Gclose(gid);
   return 0;
}

/* ---------------------------------------
 * FILL_ATTRIBUTE_NODE
 * --------------------------------------- */
int fillAttributeNode(hid_t file_id, HL_Node* node)
{
   hid_t obj=-1;
   hid_t loc_id=-1;
   hid_t type=-1, mtype=-1;
   hid_t f_space=-1;
   hsize_t* all_dims=NULL;
   hsize_t npoints;
   int ndims,i;
   char parent[512],child[512];
   unsigned char* dataptr=NULL;
   unsigned char* rawdataptr=NULL;
   char* tmpName=NULL;
   HL_Type parentType=UNDEFINED_ID;
   H5G_stat_t statbuf;

   HL_SPEWDEBUG0("ENTER: fillAttributeNode");
   if(!extractParentChildName(node,parent,child)) {
      HL_ERROR0("Failed to extract parent/child");
      return 0;
   }

   if(strcmp(parent,"")!=0) {
      disableErrorReporting(); /*Just quick hack to bypass the error reporting,
				*if failed to open a dataset/or group*/
      if((loc_id=H5Gopen(file_id,parent))<0) {
	 if((loc_id=H5Dopen(file_id,parent))>=0) {
	    parentType=DATASET_ID;
	 }
      } else {
	 parentType=GROUP_ID;
      }
      enableErrorReporting();
      if(loc_id<0) {
	 HL_ERROR2("Parent '%s' to attribute '%s' could not be opened", \
		    parent,child);
	 return 0;
      }      
   } else {
      if((loc_id=H5Gopen(file_id,"/"))<0) {
	 HL_ERROR1("Could not open root group when reading attr '%s'",
		    child);
	 return 0;
      }
      parentType=GROUP_ID;
   }

   if((obj=H5Aopen_name(loc_id,child))<0) {
      return 0;
   }

   if((type=H5Aget_type(obj))<0) {
      HL_ERROR0("Could not get attribute type");
      goto fail;
   }

   if((mtype=getFixedType(type))<0) {
      HL_ERROR0("Could not create fixed attribute type");
      goto fail;
   }

   if(H5Tget_class(mtype)==H5T_COMPOUND) {
      if(H5Tcommitted(type)>0) {
	 H5Gget_objinfo(type,".",TRUE,&statbuf);
      }
      if(!(node->compoundDescription = buildTypeDescriptionFromTypeHid(mtype))) {
	 HL_ERROR0("Failed to create compound data description for attribute");
	 goto fail;
      }
      if(H5Tcommitted(type)>0) {
	 node->compoundDescription->objno[0]=statbuf.objno[0];
	 node->compoundDescription->objno[1]=statbuf.objno[1];
      }
   }

   if((f_space=H5Aget_space(obj))>=0) {
      ndims = H5Sget_simple_extent_ndims(f_space);
      npoints = H5Sget_simple_extent_npoints(f_space);
      if(ndims>0) { 
	 if(!(all_dims=(hsize_t*)malloc(sizeof(hsize_t)*ndims))) {
	    HL_ERROR0("Could not allocate memory for node dims\n");
	    goto fail;
	 }
      }

      H5Sget_simple_extent_dims(f_space,all_dims,NULL);
      if(H5Sis_simple(f_space)>=0) {
	 node->dSize = H5Tget_size(mtype);
	 if(!(dataptr=(unsigned char*)malloc(H5Tget_size(mtype)*npoints))) {
	    HL_ERROR0("Could not allocate memory for attribute data");
	    goto fail;
	 }
	 if(H5Aread(obj,mtype,dataptr)<0) {
	    HL_ERROR0("Could not read attribute data\n");
	    goto fail;
	 }

	 /* Ahe added this 2001-01-17 */
	 node->rdSize = H5Tget_size(type);
	 if(!(rawdataptr = (unsigned char*)malloc(H5Tget_size(type)*npoints))) {
	    HL_ERROR0("Could not allocate memory for raw attribute data");
	    goto fail;
	 }
	 if(H5Aread(obj,type,rawdataptr)<0) {
	    HL_ERROR0("Could not read raw attribute data");
	    goto fail;
	 }
      } else {
	 HL_ERROR0("Attribute dataspace was not simple, can't handle");
	 goto fail;
      }
   } else {
      HL_ERROR0("Could not get dataspace for attribute");
      goto fail;
   }

   if(!(tmpName=getFormatNameString(mtype))) {
      goto fail;
   }
   strcpy(node->format,tmpName);

   if(all_dims) {
      for(i=0;i<ndims;i++)
	 node->dims[i]=all_dims[i];
      free(all_dims);
   }
   node->ndims=ndims;
   node->typeId=mtype;

   if(obj>=0) H5Aclose(obj);
   if(loc_id>=0) {
      if(parentType==GROUP_ID) {
	 H5Gclose(loc_id);
      } else if(parentType==DATASET_ID) {
	 H5Dclose(loc_id);
      } else {
	 HL_ERROR1("Could not determine type of loc_id for '%s', could not close",
		    parent);
      }
   }
   /* Mark the node as original */
   node->mark = NMARK_ORIGINAL;

   if(type>=0) H5Tclose(type);
   if(f_space>=0) H5Sclose(f_space);
   node->data=dataptr;

   node->rawdata = rawdataptr; /* Ahe added this 2001-01-17 */
   
   if(tmpName) free(tmpName);
   return 1;
fail:
   if(obj>=0) H5Aclose(obj);
   if(loc_id>=0) {
      if(parentType==GROUP_ID) {
	 H5Gclose(loc_id);
      } else if(parentType==DATASET_ID) {
	 H5Dclose(loc_id);
      } else {
	 HL_ERROR0("Could not determine type of loc_id, could not close");
      }
   }
   if(type>=0) H5Tclose(type);
   if(mtype>=0) H5Tclose(mtype);
   if(f_space>=0) H5Sclose(f_space);
   if(all_dims) free(all_dims);
   if(dataptr) free(dataptr);
   if(rawdataptr) free(rawdataptr);/* Ahe added this 2001-01-17 */
   if(tmpName) free(tmpName);

   return 0;
}

/* ---------------------------------------
 * LOCATE_NAME_FOR_REFERENCE
 * --------------------------------------- */
char* locateNameForReference(hid_t file_id, hobj_ref_t* ref)
{
   hid_t gid=-1;
   ReferenceLookup lookup;

   HL_DEBUG0("ENTER: locateNameForReference");
   
   lookup.file_id=file_id;
   lookup.ref=ref;
   strcpy(lookup.tmp_name,"");
   strcpy(lookup.found_name,"");
   lookup.reffound=0;
   
   if((gid=H5Gopen(file_id,"."))<0) {
      HL_ERROR0("Failed to open root group");
      goto fail;
   }

   H5Giterate(file_id,".",NULL,refGroupLocationIterator,&lookup);
   H5Gclose(gid);

   if(lookup.reffound==0)
      goto fail;
   
   return strdup(lookup.found_name);
 fail:
   return NULL;
}

/* ---------------------------------------
 * FILL_REFERENCE_NODE
 * --------------------------------------- */
int fillReferenceNode(hid_t file_id, HL_Node* node)
{
   hid_t obj=-1;
   hid_t loc_id=-1;
   char parent[512],child[512];
   HL_Type parentType=UNDEFINED_ID;
   hobj_ref_t ref;
   char* refername=NULL;
   char* tmpName=NULL;

   HL_DEBUG0("ENTER: fillReferenceNode");
   if(!extractParentChildName(node,parent,child)) {
      HL_ERROR0("Failed to extract parent/child");
      return 0;
   }

   if(strcmp(parent,"")!=0) {
      disableErrorReporting(); /*Just quick hack to bypass the error reporting,
				*if failed to open a dataset/or group*/
      if((loc_id=H5Gopen(file_id,parent))<0) {
	 if((loc_id=H5Dopen(file_id,parent))>=0) {
	    parentType=DATASET_ID;
	 }
      } else {
	 parentType=GROUP_ID;
      }
      enableErrorReporting();
      if(loc_id<0) {
	 HL_ERROR2("Parent '%s' to attribute '%s' could not be opened", \
		    parent,child);
	 return 0;
      }      
   } else {
      if((loc_id=H5Gopen(file_id,"/"))<0) {
	 HL_ERROR1("Could not open root group when reading attr '%s'",
		    child);
	 return 0;
      }
      parentType=GROUP_ID;
   }

   if((obj=H5Aopen_name(loc_id,child))<0) {
      goto fail;
   }
   if(H5Aread(obj,H5T_STD_REF_OBJ,&ref)<0) {
      HL_ERROR0("Could not read reference\n");
      goto fail;
   }

   if(!(refername=locateNameForReference(file_id, &ref))) {
      HL_ERROR2("WARNING: Could not locate name of object referenced by: %s/%s" \
		" will set referenced object to UNKNOWN.", parent,child);
      refername=strdup("UNKNOWN");
   }
   node->dSize = strlen(refername)+1;
   node->data=(unsigned char*)strdup(refername);
   node->rdSize = strlen(refername)+1;
   node->rawdata=(unsigned char*)strdup(refername);
   node->ndims=0;
   node->mark = NMARK_ORIGINAL;
   node->typeId=H5Tcopy(H5T_C_S1);
   H5Tset_size(node->typeId,node->dSize);
   if(!(tmpName=getFormatNameString(node->typeId))) {
      HL_ERROR0("Failed to get name string");
      goto fail;
   }
   strcpy(node->format,tmpName);
   
   if(loc_id>=0) {
      if(parentType==DATASET_ID)
	 H5Dclose(loc_id);
      else
	 H5Gclose(loc_id);
   }
   if(obj>=0)
      H5Aclose(obj);
   if(refername) free(refername);
   if(tmpName) free(tmpName);
   return 1;
 fail:
   if(loc_id>=0) {
      if(parentType==DATASET_ID)
	 H5Dclose(loc_id);
      else
	 H5Gclose(loc_id);
   }
   if(obj>=0)
      H5Aclose(obj);
   if(refername) free(refername);
   if(tmpName) free(tmpName);
   return 0;
}

/* ---------------------------------------
 * FILL_DATASET_NODE
 * --------------------------------------- */
int fillDatasetNode(hid_t file_id, HL_Node* node)
{
   hid_t obj=-1;
   hid_t type=-1;
   H5G_stat_t statbuf;
   int ndims,i;
   hsize_t* all_dims=NULL;
   hsize_t npoints;
   hid_t f_space=-1;
   hid_t mtype=-1;
   unsigned char* dataptr=NULL;
   char* tmpName=NULL;

   HL_DEBUG0("ENTER: fillDatasetNode");
   
   if((obj=H5Dopen(file_id,node->name))<0) {
      goto fail;
   }

   /* What datatype was this dataset stored as? */
   if((type=H5Dget_type(obj))<0) {
      HL_ERROR0("Failed to get type from dataset");
      goto fail;
   }

   /* What size does the type have? */
   if((f_space=H5Dget_space(obj))>0) { /*Get the space description for the dataset */
      ndims=H5Sget_simple_extent_ndims(f_space);
      npoints=H5Sget_simple_extent_npoints(f_space);
      all_dims=(hsize_t*)malloc(sizeof(hsize_t)*ndims);
      if(H5Sget_simple_extent_dims(f_space,all_dims,NULL)!=ndims) {
	 HL_ERROR0("Could not get dimensions from dataset");
	 goto fail;
      }
      /* Translate the type into a native dataspace */
      mtype = getFixedType(type);
      /*mtype = H5Tcopy(type);*/

      if(H5Tget_class(mtype)==H5T_COMPOUND) {
	 if(H5Tcommitted(type)>0) {
	    H5Gget_objinfo(type,".",TRUE,&statbuf);
	 }
	 if(!(node->compoundDescription = buildTypeDescriptionFromTypeHid(mtype))) {
	    HL_ERROR0("Failed to create compound data description for dataset");
	    goto fail;
	 }
	 if(H5Tcommitted(type)>0) {
	    node->compoundDescription->objno[0]=statbuf.objno[0];
	    node->compoundDescription->objno[1]=statbuf.objno[1];
	 }
      }

      if(H5Sis_simple(f_space)>=0) { /*Only allow simple dataspace, nothing else supported by HDF5 anyway */
	 node->dSize = H5Tget_size(mtype);
	 node->rdSize = 0;
	 dataptr=(unsigned char*)malloc(H5Tget_size(mtype)*npoints);
	 H5Sselect_all(f_space); /* Mark for selection */
	 if(H5Dread(obj,mtype,H5S_ALL,H5S_ALL,H5P_DEFAULT,dataptr)<0) {
	    HL_ERROR0("Failed to read dataset");
	    goto fail;
	 }
      } else {
	 HL_ERROR0("Dataspace for dataset was not simple, this is not supported");
	 goto fail;
      }
   } else {
      HL_ERROR0("Failure getting space description for dataset");
      goto fail;
   }

   /* Translate the type into a recognizable string, e.g. int, short, string, compound */
   if(!(tmpName=getFormatNameString(mtype))) {
      goto fail;
   }
   strcpy(node->format,tmpName);

   /* Filling dataset */
   node->ndims=ndims;
   for(i=0;i<ndims;i++) {
      node->dims[i]=all_dims[i];
   }

   /* Remember to keep the type pointer, can be hard to interpreet the data
    * later otherwise. */
   node->typeId=mtype;

   /*Better keep data also*/
   node->data=(unsigned char*)dataptr;

   /* Mark the node as original */
   node->mark = NMARK_ORIGINAL;

   if(obj>=0) H5Dclose(obj);
   if(type>=0) H5Tclose(type);
   if(all_dims) free(all_dims);
   if(f_space>=0) H5Sclose(f_space);
   if(tmpName) free(tmpName);

   return 1;
 fail:
   if(obj>=0) H5Dclose(obj);
   if(type>=0) H5Tclose(type);
   if(all_dims) free(all_dims);
   if(f_space>=0) H5Sclose(f_space);
   if(mtype>=0) H5Tclose(mtype);
   if(dataptr) free(dataptr);
   if(tmpName) free(tmpName);
   return 0;
}

/* ---------------------------------------
 * FILL_GROUP_NODE
 * --------------------------------------- */
int fillGroupNode(hid_t file_id, HL_Node* node)
{
   hid_t obj;
   HL_SPEWDEBUG0("ENTER: fillGroupNode");
   if((obj=H5Gopen(file_id,node->name))<0) {
      return 0;
   }

   /* Mark the node as original */
   node->mark = NMARK_ORIGINAL;

   H5Gclose(obj);
   return 1;
}

HL_CompoundTypeDescription* buildTypeDescriptionFromTypeHid(hid_t type_id)
{
   HL_CompoundTypeDescription* typelist=NULL;
   HL_CompoundTypeAttribute* typenode;
   int i,j,nmembers;
   char* fname=NULL;
   hid_t mtype=-1;
   hid_t fixedType=-1;
   size_t* dims=NULL;
   hsize_t* dims_h514=NULL;
   size_t dSize;
   int ndims;
   char* tmpName=NULL;
   HL_DEBUG0("ENTER: buildTypeDescriptionFromTypeHid");
   if(!(typelist=newHL_CompoundTypeDescription())) {
      HL_ERROR0("Failed to create datatype nodelist");
      goto fail;
   }

   typelist->size=H5Tget_size(type_id);
   nmembers = H5Tget_nmembers(type_id);
   for(i=0;i<nmembers;i++) {
      size_t offset = H5Tget_member_offset(type_id,i);
      fname = H5Tget_member_name(type_id,i);
      mtype = H5Tget_member_type(type_id,i);
      /* IF MEMBER IS AN ARRAY DO SOMETHING SPECIAL!!!	*/
      /*Added by AHE 000315*/

      if((fixedType = getFixedType(mtype))<0) {
	 HL_ERROR0("Failed to convert to fixed type");
	 goto fail;
      }
      
      if(!(tmpName=getFormatNameString(fixedType))) {
	 goto fail;
      }
      
#ifdef H5_VERS_MINOR
/*#if H5_VERS_MINOR>=4*/
#if H5_VERS_MAJOR>1 || (H5_VERS_MAJOR<2 && H5_VERS_MINOR>=4)
      if (H5T_ARRAY==H5Tget_member_class(type_id,i)) {
	 ndims = H5Tget_array_ndims(mtype);
	 dims_h514 = (hsize_t*)malloc(sizeof(hsize_t)*ndims);
	 if(H5Tget_array_dims(mtype,dims_h514,NULL) < 0) {
	    HL_ERROR0("Failed to retrieve array dimensions");
	 }
	 dims = (size_t*)malloc(sizeof(size_t)*ndims);	 
	 for(j=0; j < ndims; j++) {
	    dims[j] = (size_t) dims_h514[j];
	 }
      }
      else {
	 ndims=1;
	 dims = (size_t*)malloc(sizeof(size_t)*ndims);
	 dims[0]=1;
      }	       
#else      
      ndims = H5Tget_member_dims(type_id,i,NULL,NULL);
      dims = (size_t*)malloc(sizeof(size_t)*ndims);
      ndims = H5Tget_member_dims(type_id,i,dims,NULL);
      /*
      Incompatible with version 1.4 and higher. Adam Dybbroe, 2001-04-26
      */
#endif
#else
      /* If H5_VERS_MINOR is not defined we assume version 1.4 style...	*/
      /* Previously empty here, have added the 1.4+ code here, with the hopes
       * that it is correct, printout a warning though that H5_VERS_MINOR not
       * is defined so that it is possible to see that it might be something wrong
       * here. 2003-01-16 AHE
       */
      HL_INFO0("H5_VERS_MINOR was not defined, asuming 1.4+ style should be used for retrieving dimensions\n");
      if (H5T_ARRAY==H5Tget_member_class(type_id,i)) {
	 ndims = H5Tget_array_ndims(mtype);
	 dims_h514 = (hsize_t*)malloc(sizeof(hsize_t)*ndims);
	 if(H5Tget_array_dims(mtype,dims_h514,NULL) < 0) {
	    HL_ERROR0("Failed to retrieve array dimensions");
	 }
	 dims = (size_t*)malloc(sizeof(size_t)*ndims);	 
	 for(j=0; j < ndims; j++) {
	    dims[j] = (size_t) dims_h514[j];
	 }
      }
      else {
	 ndims=1;
	 dims = (size_t*)malloc(sizeof(size_t)*ndims);
	 dims[0]=1;
      }	       
#endif
      dSize = H5Tget_size(fixedType);

      if(fname==NULL) {
	 HL_ERROR0("fname is NULL, cant use it to create CompoundTypeAttribute\n");
	 goto fail;
      }
      typenode=newHL_CompoundTypeAttribute(fname,offset,tmpName,dSize,ndims,dims);
      if(!addCompoundTypeAttribute(typelist,typenode))
	 goto fail;
      if(dims) {
	 free(dims);
	 dims=NULL;
      }
      if(dims_h514) {
	 free(dims_h514);
	 dims_h514=NULL;
      }

      if(mtype>=0)
	 H5Tclose(mtype);
      mtype=-1;

      if(fname)
	 free(fname);
      fname=NULL;
      if(tmpName) {
	 free(tmpName);
	 tmpName=NULL;
      }
      if(fixedType>=0)
	 H5Tclose(fixedType);
      fixedType=-1;
   }

   if(dims_h514) free(dims_h514);
   if(dims) free(dims);
   if(mtype>=0) H5Tclose(mtype);
   if(fixedType>=0)
      H5Tclose(fixedType);
   if(fname) free(fname);
   if(tmpName) free(tmpName);
   return typelist;
fail:
   if(dims) free(dims);
   if(dims_h514) free(dims_h514);
   if(mtype>=0) H5Tclose(mtype);
   if(fixedType>=0)
      H5Tclose(fixedType);
   if(fname) free(fname);
   if(tmpName) free(tmpName);
   freeHL_CompoundTypeDescription(typelist);
   return NULL;
}

/* ---------------------------------------
 * FILL_TYPE_NODE
 * --------------------------------------- */
int fillTypeNode(hid_t file_id, HL_Node* node)
{
   hid_t obj=-1;
   HL_CompoundTypeDescription* typelist=NULL;
   H5G_stat_t statbuf;
   HL_DEBUG0("ENTER: fillTypeNode");

   if((obj=H5Topen(file_id,node->name))<0) {
      return 0;
   }

   H5Gget_objinfo(obj,".",TRUE,&statbuf);

   if(!(typelist=buildTypeDescriptionFromTypeHid(obj))) {
      HL_ERROR0("Failed to create datatype nodelist");
      goto fail;
   }
   strcpy(typelist->hltypename,node->name);
   typelist->objno[0]=statbuf.objno[0];
   typelist->objno[1]=statbuf.objno[1];

   node->compoundDescription=typelist;

   /* Mark the node as original */
   node->mark = NMARK_ORIGINAL;

   node->hdfId=obj; /*Save hdfid for later use, which means that obj not should be closed now. */
   return 1;
fail:
   if(obj>=0) H5Tclose(obj);
   freeHL_CompoundTypeDescription(typelist);
   return 0;
}

/* ---------------------------------------
 * FILL_NODE_WITH_DATA
 * --------------------------------------- */
int fillNodeWithData(hid_t file_id, HL_Node* node)
{
   HL_SPEWDEBUG0("ENTER: fillNodeWithData");
   switch(node->type) {
   case ATTRIBUTE_ID:
      return fillAttributeNode(file_id, node);
   case DATASET_ID:
      return fillDatasetNode(file_id, node);
   case GROUP_ID:
      return fillGroupNode(file_id,node);
   case TYPE_ID:
      return fillTypeNode(file_id,node);
   case REFERENCE_ID:
      return fillReferenceNode(file_id,node);
   default:
      HL_ERROR1("Can't handle other nodetypes but '%d'",node->type);
      break;
   }
   return 0;
}
/* ---------------------------------------
 * Static methods used internally defined below
 * this comment.
 * ---------------------------------------*/

/* ---------------------------------------
 * CHECK_IF_REFERENCE_MATCH
 * Creates a reference to path and then
 * verifies if this reference is the same
 * as ref.
 * ---------------------------------------*/
static int checkIfReferenceMatch(hid_t loc_id, char* path, hobj_ref_t* ref)
{
   hobj_ref_t matchref;
   herr_t status;
   HL_DEBUG0("ENTER:  checkIfReferenceMatch");
   status = H5Rcreate(&matchref, loc_id, path, H5R_OBJECT, -1);
   if(status<0) {
      HL_ERROR1("Could not create reference to '%s'",path);
      return 0;
   }
   if(memcmp(ref,&matchref,sizeof(hobj_ref_t))==0) {
      return 1;
   }
   return 0;
}

/* ---------------------------------------
 * REF_ATTRIBUTE_LOCATION_ITERATOR
 * Iterator function for checking attribute
 * references. This does not seem to work.
 * It seems like there is no possibility
 * to create references to attributes.
 * ---------------------------------------*/
static herr_t refAttributeLocationIterator(hid_t loc_id, const char* name, void* op_data)
{
   ReferenceLookup* lookup=(ReferenceLookup*)op_data;
   char tmp1[512],tmp2[512];
   HL_DEBUG0("ENTER: refAttributeLocationIterator");
   strcpy(tmp1,lookup->tmp_name);
   sprintf(tmp2,"%s/%s",lookup->tmp_name,name);
   strcpy(lookup->tmp_name,tmp2);

   if(checkIfReferenceMatch(loc_id, (char*)name, lookup->ref)==1) {
      strcpy(lookup->found_name,lookup->tmp_name);
      lookup->reffound=1;
   }
   
   strcpy(lookup->tmp_name,tmp1);
   return 0;
}

/* ---------------------------------------
 * REF_GROUP_LOCATION_ITERATOR
 * Iterator function for checking group
 * references. 
 * ---------------------------------------*/
static herr_t refGroupLocationIterator(hid_t gid, const char* name, void* op_data)
{
   ReferenceLookup* lookup=(ReferenceLookup*)op_data;
   char tmp1[512],tmp2[512];
   H5G_stat_t statbuf;
   hid_t obj=-1;
   HL_DEBUG0("ENTER: refGroupLocationIterator");
   H5Gget_objinfo(gid,name,0,&statbuf);
   
   strcpy(tmp1,lookup->tmp_name);
   
   switch(statbuf.type) {
   case H5G_GROUP:
      if((obj=H5Gopen(gid,name))>=0) {
	 sprintf(tmp2,"%s/%s",lookup->tmp_name,name);
	 strcpy(lookup->tmp_name,tmp2);
	 if(checkIfReferenceMatch(lookup->file_id, lookup->tmp_name, lookup->ref)==1) {
	    strcpy(lookup->found_name,lookup->tmp_name);
	    lookup->reffound=1;
	 }
	 /*H5Aiterate(obj,NULL,refAttributeLocationIterator, lookup);*/
	 H5Giterate(obj,".", NULL, refGroupLocationIterator, lookup);
	 strcpy(lookup->tmp_name,tmp1);
	 H5Gclose(obj);
      }
      break;
   case H5G_DATASET:
      if((obj=H5Dopen(gid,name))>=0) {
	 sprintf(tmp2,"%s/%s",lookup->tmp_name,name);
	 strcpy(lookup->tmp_name,tmp2);
	 if(checkIfReferenceMatch(lookup->file_id, lookup->tmp_name, lookup->ref)==1) {
	    strcpy(lookup->found_name,lookup->tmp_name);
	    lookup->reffound=1;
	 }
	 /*H5Aiterate(obj,NULL,refAttributeLocationIterator,lookup);*/
	 strcpy(lookup->tmp_name,tmp1);
	 H5Dclose(obj);
      }
      break;
   case H5G_TYPE:
      sprintf(tmp2,"%s/%s",lookup->tmp_name,name);
      strcpy(lookup->tmp_name,tmp2);
      if(checkIfReferenceMatch(lookup->file_id, lookup->tmp_name, lookup->ref)==1) {
	 strcpy(lookup->found_name,lookup->tmp_name);
	 lookup->reffound=1;
      }
      strcpy(lookup->tmp_name,tmp1);
      break;
   default:
      HL_ERROR1("Undefined type for %s",name);
      HL_ERROR1("Name: %s",tmp1);
      HL_ERROR1("Type id %d",statbuf.type);
      break;
   }
   return 0;
}
