/* --------------------------------------------------------------------------
$Id: write_hlhdf.c,v 1.4 2004/03/18 11:53:50 ahenja Exp $
Program:        

Description:    Common functions for writing an HDF5 file.

Author(s):      Anders Henja

Copyright:	Swedish Meteorological and Hydrological Institute, 2000

History:
$Log: write_hlhdf.c,v $
Revision 1.4  2004/03/18 11:53:50  ahenja
Modified with support for szip, also changed so that it is possible
to specify HL_FileCreationProperty when creating a file. Also changed
how the compression handling is working. Added a nicer way to
perform debugging and error handling. Now it is possible to define
an own error handler and hook it into the system, if this is done
all error messages will be sent to the given function.

Revision 1.3  2002/12/19 13:29:14  ahenja
Added some typecastings when mallocing memory.

Revision 1.2  2002/08/06 10:29:18  ahenja
Corrected bug in write_hlhdf.c function addAttributeToCompoundType, an illegal
cast was performed for HDF5 Versions >= 1.4.x.

Revision 1.1.1.1  2001/12/14 12:02:38  ahenja
First CVS version of hlhdf

Revision 1.6  2001/12/07 13:20:32  ahenja
Updated with possibilities to read and write references.

Revision 1.5  2001/11/09 08:30:56  ahenja
Updated with possibility to write references, unfortunately i havent got the time to implement read support.

Revision 1.4  2001/11/08 12:27:20  adybbroe
Updated.

Revision 1.3  2001/04/26 14:16:19  adybbroe
Solved for version 1.4 incompatibility.

Revision 1.2  2000/03/22 08:41:33  ahenja
Updated with spell checking and header information.

-----------------------------------------------------------------------------*/
#include "hlhdf.h"
#include "hlhdf_debug.h"

/************************************************
 * commitType
 ***********************************************/
herr_t commitType(hid_t loc_id, const char* name, hid_t type_id)
{
   HL_DEBUG0("ENTER: commitType");
   return H5Tcommit(loc_id,name,type_id);
}

/************************************************
 * createStringType
 ***********************************************/
hid_t createStringType(size_t length)
{
   hid_t type;
   HL_DEBUG0("ENTER: createStringType");
   type=H5Tcopy(H5T_C_S1);
   H5Tset_size(type,length);
   return type;
}
/************************************************
 * setTypeSize
 ***********************************************/
herr_t setTypeSize(hid_t type_id,size_t theSize)
{
   HL_DEBUG0("ENTER: setTypeSize");
   return H5Tset_size(type_id,theSize);
}

/************************************************
 * closeType
 ***********************************************/
herr_t closeType(hid_t type_id)
{
   HL_DEBUG0("ENTER: closeType");
   return H5Tclose(type_id);
}

/************************************************
 * writeScalarDataAttribute
 ***********************************************/
herr_t writeScalarDataAttribute(hid_t loc_id, hid_t type_id, const char* name, void* buf)
{
   hid_t aid=-1;
   hid_t attr_id=-1;
   herr_t status = 0;
   HL_SPEWDEBUG0("ENTER: writeScalarDataAttribute");
   if((aid = H5Screate(H5S_SCALAR))<0) {
      HL_ERROR0("Failed to create scalar data space");
      status = -1;
      goto done;
   }
   if((attr_id=H5Acreate(loc_id,name,type_id,aid,H5P_DEFAULT))<0) {
      HL_ERROR0("Failed to create scalar attribute");
      status = -1;
      goto done;
   }

   if(H5Awrite(attr_id,type_id,buf)<0) {
      HL_ERROR0("Failed to write scalar data to file");
      status = -1;
   }

 done:
   if(aid!=-1)
      H5Sclose(aid);
   if(attr_id!=-1)
      H5Aclose(attr_id);
   return status;
}

/************************************************
 * writeScalarDataAttribute_fmt
 ***********************************************/
herr_t writeScalarDataAttribute_fmt(hid_t loc_id, const char* fmt, const char* name, void* buf)
{
   hid_t typeId=-1;
   herr_t status=0;
   HL_DEBUG0("ENTER: writeScalarDataAttribute_fmt");
   if((typeId=translateCharToDatatype(fmt))<0) {
      HL_ERROR0("Failed to translate from format to HDF type");
      return -1;
   }
   status = writeScalarDataAttribute(loc_id,typeId,name,buf);
   H5Tclose(typeId);
   return status;
}

/************************************************
 * writeSimpleDataAttribute
 ***********************************************/
herr_t writeSimpleDataAttribute(hid_t loc_id, hid_t type_id, const char* name, int ndims, hsize_t* dims, void* buf)
{
   hid_t attr_id=-1;
   hid_t dataspace=-1;
   herr_t status = 0;
   HL_DEBUG0("ENTER: writeSimpleDataAttribute");
   if((dataspace = H5Screate_simple(ndims,dims,NULL))<0) {
      HL_ERROR0("Failed to create simple dataspace for attribute");
      status = -1;
      goto done;
   }

   if((attr_id=H5Acreate(loc_id,name,type_id,dataspace,H5P_DEFAULT))<0) {
      HL_ERROR0("Failed to create simple attribute");
      status = -1;
      goto done;
   }

   if(H5Awrite(attr_id,type_id,buf)<0) {
      HL_ERROR0("Failed to write simple data attribute to file");
      status = -1;
   }

 done:
   if(dataspace!=-1)
      H5Sclose(dataspace);
   if(attr_id!=-1)
      H5Aclose(attr_id);
   return status;
}

/************************************************
 * writeSimpleDataAttribute_fmt
 ***********************************************/
herr_t writeSimpleDataAttribute_fmt(hid_t loc_id, const char* fmt, const char* name,
				    int ndims, hsize_t* dims, void* buf)
{
   hid_t typeId=-1;
   herr_t status=0;
   HL_DEBUG0("ENTER: writeSimpleDataAttribute_fmt");
   if((typeId=translateCharToDatatype(fmt))<0) {
      HL_ERROR0("Failed to translate from format to HDF type");
      return -1;
   }
   status = writeSimpleDataAttribute(loc_id,typeId,name,ndims,dims,buf);
   H5Tclose(typeId);
   return status;
}

/************************************************
 * createSimpleHLDataset
 ***********************************************/
hid_t createSimpleHLDataset(hid_t loc_id, hid_t type_id, const char* name, int ndims, hsize_t* dims,
			    void* buf,HL_Compression* compress)
{
   hid_t dataset=-1;
   hid_t dataspace=-1;
   hid_t props=-1;
   HL_SPEWDEBUG0("ENTER: createSimpleHLDataset");
   if((dataspace=H5Screate_simple(ndims,dims,NULL))<0) {
      HL_ERROR0("Failed to create simple dataspace for dataset");
      goto done;
   }

   if(compress!=NULL &&
      (compress->type==CT_SZLIB || (compress->type=CT_ZLIB &&
				    (compress->level>0 || compress->level<=9) ) ) ) {

      if((props = H5Pcreate(H5P_DATASET_CREATE))<0) {
	 HL_ERROR0("Failed to create the compression property");
	 goto done;
      }
      if(H5Pset_chunk(props,ndims,dims)<0) {
	 HL_ERROR0("Failed to set chunk size");
	 goto done;
      }
      if(compress->type==CT_ZLIB) {
	 if(H5Pset_deflate(props,compress->level)<0) {
	    HL_ERROR1("Failed to set z compression to level %d",compress->level);
	    goto done;
	 }
      } else {
	 if(H5Pset_szip(props, compress->szlib_mask, compress->szlib_px_per_block)<0) {
	    HL_ERROR2("Failed to set the szip compression, mask=%d, px_per_block=%d",\
		      compress->szlib_mask, compress->szlib_px_per_block);
	    goto done;
	 }
      }

      if((dataset=H5Dcreate(loc_id,name,type_id,dataspace,props))<0) {
	 HL_ERROR0("Failed to create the dataset");
	 goto done;
      }
   } else {
      if((dataset=H5Dcreate(loc_id,name,type_id,dataspace,H5P_DEFAULT))<0) {
	 HL_ERROR0("Failed to create the dataset");
	 goto done;
      }
   }

   if(buf!=NULL) {
      if(H5Dwrite(dataset,type_id,H5S_ALL,H5S_ALL,H5P_DEFAULT,buf)<0) {
	 HL_ERROR0("Failed to write dataset");
	 goto done;
      }
   }

 done:
   if(dataspace!=-1)
      H5Sclose(dataspace);
   if(props!=-1)
      H5Pclose(props);

   return dataset;   
}

/************************************************
 * createSimpleDataset
 ***********************************************/
hid_t createSimpleDataset(hid_t loc_id, hid_t type_id, const char* name, int ndims, hsize_t* dims, void* buf,int compress)
{

   HL_Compression theCompression;
   HL_DEBUG0("ENTER: createSimpleDataset");
   initHL_Compression(&theCompression,CT_ZLIB);
   theCompression.level = compress;
   return createSimpleHLDataset(loc_id,type_id,name,ndims,dims,buf,&theCompression);
}

/************************************************
 * createSimpleDataset_fmt
 ***********************************************/
hid_t createSimpleHLDataset_fmt(hid_t loc_id, const char* fmt, const char* name, int ndims, hsize_t* dims,
				void* buf, HL_Compression* compress)
{
   hid_t typeId=-1;
   hid_t dset=0;
   HL_DEBUG0("ENTER: createSimpleHLDataset_fmt");
   if((typeId=translateCharToDatatype(fmt))<0) {
      HL_ERROR0("Failed to translate from format to HDF type");
      return -1;
   }
   dset = createSimpleHLDataset(loc_id,typeId,name,ndims,dims,buf,compress);
   H5Tclose(typeId);
   return dset;
}

/************************************************
 * createSimpleDataset_fmt
 ***********************************************/
hid_t createSimpleDataset_fmt(hid_t loc_id, const char* fmt, const char* name, int ndims, hsize_t* dims, void* buf, int compress)
{
   HL_Compression theCompression;
   HL_DEBUG0("ENTER: createSimpleDataset_fmt");
   initHL_Compression(&theCompression,CT_ZLIB);
   theCompression.level = compress;
   return createSimpleHLDataset_fmt(loc_id,fmt,name,ndims,dims,buf,&theCompression);
}

/************************************************
 * closeDataset
 ***********************************************/
herr_t closeDataset(hid_t loc_id)
{
   HL_DEBUG0("ENTER: closeDataset");
   return H5Dclose(loc_id);
}


/************************************************
 * createCompoundType
 ***********************************************/
hid_t createCompoundType(size_t size)
{
   HL_DEBUG0("ENTER: createCompoundType");
   return H5Tcreate(H5T_COMPOUND,size);
}

/************************************************
 * addAttributeToCompoundType
 ***********************************************/
herr_t addAttributeToCompoundType(hid_t loc_id, const char* name, size_t offset,hid_t type_id)
{
   HL_DEBUG0("ENTER: addAttributeToCompoundType");
   return H5Tinsert(loc_id,name,offset,type_id);
}

/************************************************
 * addAttributeToCompoundType_fmt
 ***********************************************/
herr_t addAttributeToCompoundType_fmt(hid_t loc_id, const char* name, size_t offset,const char* fmt)
{
   herr_t status;
   hid_t type_id = translateCharToDatatype(fmt);
   HL_DEBUG0("ENTER: addAttributeToCompoundType_fmt");
   if(type_id<0)
      return -1;
   status = addAttributeToCompoundType(loc_id,name,offset,type_id);
   H5Tclose(type_id);
   return status;
}

/************************************************
 * addArrayToCompoundType
 ***********************************************/
herr_t addArrayToCompoundType(hid_t loc_id, const char* name, size_t offset, int ndims,size_t* dims,hid_t type_id)
{
   herr_t status;
   hid_t array_type;
   int i;
   hsize_t* dims_hsize_t=(hsize_t*)malloc(sizeof(hsize_t)*ndims);
   HL_DEBUG0("ENTER: addArrayToCompoundType");
   if(!dims_hsize_t) {
      HL_ERROR0("Failed to allocate memory for temporary hsize_t dims");
      return -1;
   }
   for(i=0;i<ndims;i++) {
      dims_hsize_t[i]=dims[i];
   }
#ifdef H5_VERS_MINOR
#if H5_VERS_MAJOR>1 || (H5_VERS_MAJOR<2 && H5_VERS_MINOR>=4)
   array_type=H5Tarray_create(type_id,ndims,dims_hsize_t,NULL);
   /*array_type=H5Tarray_create(type_id,ndims,(hsize_t*)dims,NULL);*/
   status=H5Tinsert(loc_id,name,offset,array_type);
   H5Tclose(array_type);
#else
   status=H5Tinsert_array(loc_id,name,offset,ndims,dims,NULL,type_id);
   /*
     Incompatible with version 1.4 and later. Adam Dybbroe, 2001-04-26
     The below should be in line with the latest version...
   */     
#endif
#else
   array_type=H5Tarray_create(type_id,ndims,dims_hsize_t,NULL);
   status=H5Tinsert(loc_id,name,offset,array_type);
   H5Tclose(array_type);
#endif
   free(dims_hsize_t);
   return status;
}

/************************************************
 * addArrayToCompoundType_fmt
 ***********************************************/
herr_t addArrayToCompoundType_fmt(hid_t loc_id, const char* name, size_t offset, \
				  int ndims, size_t* dims, const char* fmt)
{
   hid_t type_id;
   herr_t status;
   HL_DEBUG0("ENTER: addArrayToCompoundType_fmt");
   type_id = translateCharToDatatype(fmt); 
   if(type_id<0)
      return -1;
   status = addArrayToCompoundType(loc_id,name,offset,ndims,dims,type_id);
   H5Tclose(type_id);

   return status;
}


/************************************************
 * createGroup
 ***********************************************/
hid_t createGroup(hid_t loc_id, const char* groupname,const char* comment)
{
   hid_t group_id;
   HL_DEBUG0("ENTER: createGroup");
   if((group_id=H5Gcreate(loc_id,groupname,0))<0) {
      HL_ERROR0("Failed to create group");
      return -1;
   }
   if(comment) {
      if(H5Gset_comment(loc_id,groupname,comment)<0) {
	 HL_ERROR0("Failed to set comment for group");
      }
   }
   return group_id;
}

/************************************************
 * closeGroup
 ***********************************************/
herr_t closeGroup(hid_t group_id)
{
   HL_DEBUG0("ENTER: closeGroup");
   return H5Gclose(group_id);
}

/************************************************
 * createReference
 ***********************************************/
herr_t createReference(hid_t loc_id, hid_t file_id, const char* name,
		       const char* targetname)
{
   hid_t aid=-1;
   hid_t attr_id=-1;
   hid_t attr_type=-1;
   hobj_ref_t  ref;
   herr_t status = 0;
   HL_DEBUG0("ENTER: createReference");
   if((aid = H5Screate(H5S_SCALAR))<0) {
      HL_ERROR0("Failed to create scalar data space");
      status = -1;
      goto done;
   }

   if((attr_type=H5Tcopy(H5T_STD_REF_OBJ))<0) {
      HL_ERROR0("Failed to copy H5T reference type");
      status = -1;
      goto done;
   }
   
   if((attr_id=H5Acreate(loc_id,name,attr_type,aid,H5P_DEFAULT))<0) {
      HL_ERROR0("Failed to create scalar attribute");
      status = -1;
      goto done;
   }

   /* Of course I assume that the dataset is accurately specified
    * by targetname, otherwise well, implement a check here, will
    * probably require that the in argument is the actual dataset id :-/
    */
   if(H5Rcreate(&ref, file_id, targetname, H5R_OBJECT, -1)<0) {
      HL_ERROR0("Failed to create reference object");
      status = -1;
      goto done;
   }
   if(H5Awrite(attr_id,attr_type,&ref)<0) {
      HL_ERROR0("Failed to write scalar data to file");
      status = -1;
   }

 done:
   if(aid!=-1)
      H5Sclose(aid);
   if(attr_id!=-1)
      H5Aclose(attr_id);
   if(attr_type!=-1)
      H5Tclose(attr_type);
   return status;
}
