/* --------------------------------------------------------------------------
$Id: pyhl.c,v 1.12 2008/03/25 09:26:41 adybbroe Exp $
Program:        

Description:    The pyhl module source code, provides a python interface
		to the HighLevel protocol for writing and reading
		HDF5 files.

Author(s):      Anders Henja

Copyright:	Swedish Meteorological and Hydrological Institute, 2000

History:
$Log: pyhl.c,v $
Revision 1.12  2008/03/25 09:26:41  adybbroe
Tidy up to avoid warnings during compilation-

Revision 1.11  2007/11/27 08:08:01  adybbroe
Replaced PyMem_DEL with PyObject_Del, to be compatible with python 2.5
and beyond. From Python 2.5.1 release-notes
(http://www.python.org/download/releases/2.5.1/NEWS.txt):
PyMem_DEL no longer maps to PyObject_Free. It maps to the system
``free()`` now.  If memory is obtained via the ``PyObject_`` family,
it must be released via the ``PyObject_`` family, and likewise for the
``PyMem_`` family.

Revision 1.10  2006/02/02 15:59:36  dmichels
Added selectMetadata method, based on the selectMetadataNodes function
in read_vhlhdf.c

Revision 1.9  2004/10/04 12:51:48  ahenja
Updated pyhl.c with correction regarding compound_data.
Seems like it couldnt handle if ndims=1 && dims[0]==1 as
scalar!?

Revision 1.8  2004/09/27 12:11:08  ahenja
Removed coredump when trying to fetch rawdata even though
none has been read.

Revision 1.7  2004/07/01 09:28:18  ahenja
Updated with support for new compression and new loghandling

Revision 1.6  2004/03/29 14:58:24  ahenja
Updated with support for turning on/off outputs from
both HLHDF and HDF5 all depending on needs, also added
so that it is possible to reroute error messages by
writing client specific functions that handles the
various messages.

Revision 1.5  2004/03/18 11:53:52  ahenja
Modified with support for szip, also changed so that it is possible
to specify HL_FileCreationProperty when creating a file. Also changed
how the compression handling is working. Added a nicer way to
perform debugging and error handling. Now it is possible to define
an own error handler and hook it into the system, if this is done
all error messages will be sent to the given function.

Revision 1.4  2003/02/04 13:04:58  adybbroe
Corrected for potential memory leakage in pyhl.c.
Bug found and reported by Fredrik Lundh.

Revision 1.3  2003/01/15 11:57:18  ahenja
Added support for allowing both old and new numeric to be used.

Revision 1.2  2002/06/04 06:57:17  adybbroe
*** empty log message ***

Revision 1.1.1.1  2001/12/14 12:02:42  ahenja
First CVS version of hlhdf

Revision 1.7  2001/11/09 11:53:16  adybbroe
removed printf statement.

Revision 1.6  2001/11/08 12:07:14  ahenja
Basically done nothing, just some cleanup.

 * Revision 1.5  2001/01/18  12:47:49  ahenja
 * Added lots of changes.
 *
Revision 1.4  2000/08/21 08:44:30  dmichels
Replaced PyArray_FromDimsAndData and
removed trailing \OOO character when reading strings.

Revision 1.3  2000/04/11 10:08:41  ahenja
Added functionality for handle lists of ints, floats, doubles, longs and strings.

Revision 1.2  2000/03/22 08:59:06  ahenja
Updated with spell checking and header information.
	
-----------------------------------------------------------------------------*/
#include <Python.h>
/* You have to define HLHDF_PYMODULE_WITH_IMPORT_ARRAY to be able to force the api to be correctly set here */
#define HLHDF_PYMODULE_WITH_IMPORT_ARRAY
#include "pyhlhdf_common.h"  /* this includes arrayobject.h */
#include "read_vhlhdf.h"
#include "write_hlhdf.h"
#include "write_vhlhdf.h"
#include "update_vhlhdf.h"
#include "hlhdf_debug.h"

static PyObject *ErrorObject;

/* Some defines for error handling */
#define raiseException(type,msg)\
{PyErr_SetString(type,msg);return NULL;}

#define setException(type,msg)\
{PyErr_SetString(type,msg);}

typedef struct {
   PyObject_HEAD /*Always have to be on top*/
   HL_NodeList* nodelist;
} PyhlObject;

typedef struct {
   PyObject_HEAD /*Always have to be on top*/
   HL_Node* node;
} PyhlNode;

typedef struct {
   PyObject_HEAD /*Always have to be on top*/
   HL_FileCreationProperty* props;
} PyhlFileCreationProperty;

typedef struct {
   PyObject_HEAD /*Always have to be on top*/
   HL_Compression* compr;
} PyhlCompression;

/*
 *Define that the PyhlObject is of the type Pyhl_Type
 */
staticforward PyTypeObject Pyhl_Type;
staticforward PyTypeObject PyhlNode_Type;
staticforward PyTypeObject PyhlFileCreationProperty_Type;
staticforward PyTypeObject PyhlCompression_Type;

#define Pyhl_Check(op) ((op)->ob_type == &Pyhl_Type)
#define PyhlNode_Check(op) ((op)->ob_type == &PyhlNode_Type)
#define PyhlFileCreationProperty_Check(op) ((op)->ob_type == &PyhlFileCreationProperty_Type)
#define PyhlCompression_Check(op) ((op)->ob_type == &PyhlCompression_Type)

/* Pyhl module functions */
static void _dealloc(PyhlObject* pyhl_object)
{
   /*Nothing yet*/
   if(!pyhl_object)
      return;

   freeHL_NodeList(pyhl_object->nodelist);

   /*	PyMem_DEL(pyhl_object);
     PyMem_DEL no longer maps to PyObject_Free.
     It maps to the system ``free()`` now.  If memory
     is obtained via the ``PyObject_`` family, it must be released via the
     ``PyObject_`` family, and likewise for the ``PyMem_`` family.
     Adam Dybbroe, 2007-10-14
   */
   PyObject_Del(pyhl_object);
   
}

static void _dealloc_pyhlnode(PyhlNode* pyhl_node)
{
   /*Nothing yet*/
   if(!pyhl_node)
      return;

   freeHL_Node(pyhl_node->node);

   /*	PyMem_DEL(pyhl_node);
     PyMem_DEL no longer maps to PyObject_Free.
     It maps to the system ``free()`` now.  If memory
     is obtained via the ``PyObject_`` family, it must be released via the
     ``PyObject_`` family, and likewise for the ``PyMem_`` family.
     Adam Dybbroe, 2007-10-14
   */
   PyObject_Del(pyhl_node);

   return;
}

static void _dealloc_pyhlfilecreationproperty(PyhlFileCreationProperty* val)
{
   if(!val)
      return;
   freeHL_fileCreationProperty(val->props);
   /*	PyMem_DEL(val);
     PyMem_DEL no longer maps to PyObject_Free.
     It maps to the system ``free()`` now.  If memory
     is obtained via the ``PyObject_`` family, it must be released via the
     ``PyObject_`` family, and likewise for the ``PyMem_`` family.
     Adam Dybbroe, 2007-10-14
   */
   PyObject_Del(val);

}

static void _dealloc_pyhlcompression(PyhlCompression* val)
{
   if(!val)
      return;
   freeHL_Compression(val->compr);
   /*	PyMem_DEL(val);
     PyMem_DEL no longer maps to PyObject_Free.
     It maps to the system ``free()`` now.  If memory
     is obtained via the ``PyObject_`` family, it must be released via the
     ``PyObject_`` family, and likewise for the ``PyMem_`` family.
     Adam Dybbroe, 2007-10-14
   */
   PyObject_Del(val);
}

static PyObject* _pyhl_new_nodelist(PyObject* self, PyObject* args)
{
   PyhlObject* retv=NULL;
   retv=PyObject_NEW(PyhlObject,&Pyhl_Type);
   if(!retv)
      return NULL;

   if(!(retv->nodelist=newHL_NodeList())) {
      setException(PyExc_MemoryError,"Failed to create HL NodeList\n");
      _dealloc(retv);
      retv=NULL;
   }

   return (PyObject*)retv;
}

static PyObject* _pyhl_new_node(PyObject* self, PyObject* args)
{
   PyhlNode* retv=NULL;
   PyObject* comprinobj=NULL;
   PyhlCompression* compr=NULL;
   HL_Type type;
   char* nodename;

   if(!PyArg_ParseTuple(args,"is|O",&type,&nodename,&comprinobj))
      return NULL;

   switch(type) {
   case ATTRIBUTE_ID:
   case GROUP_ID:
   case DATASET_ID:
   case TYPE_ID:
   case REFERENCE_ID:
      break;
   default:
      printf("Unsupported node type %d\n",type);
      return NULL;
   }

   if(comprinobj!=NULL) {
      if(!PyhlCompression_Check(comprinobj)) {
	 setException(PyExc_AttributeError,"Third argument must be of type PyhlCompression");
	 return NULL;
      }
      if(type!=DATASET_ID) {
	 setException(PyExc_AttributeError,"It is only possible to specify PyhlCompression when creating a node of type DATASET_ID");
	 return NULL;
      }
      compr =(PyhlCompression*)comprinobj;
   }
   
   retv=PyObject_NEW(PyhlNode,&PyhlNode_Type);
   if(!retv)
      return NULL;

   switch(type) {
   case ATTRIBUTE_ID:
      retv->node = newHL_Attribute(nodename);
      break;
   case GROUP_ID:
      retv->node = newHL_Group(nodename);
      break;
   case DATASET_ID:
      retv->node = newHL_Dataset(nodename);
      break;
   case TYPE_ID:
      retv->node = newHL_Datatype(nodename);
      break;
   case REFERENCE_ID:
      retv->node = newHL_Reference(nodename);
      break;
   default:
      retv->node = NULL;
      break;
   }

   if(!(retv->node)) {
      setException(PyExc_MemoryError,"Failed to create Node\n");
      _dealloc_pyhlnode(retv);
      retv=NULL;
   }

   retv->node->compression=dupHL_Compression((compr!=NULL)?compr->compr:NULL);
   return (PyObject*)retv;
}

static PyObject* _pyhl_new_filecreationproperty(PyObject* self, PyObject* args)
{
   PyhlFileCreationProperty* retv=NULL;
   retv=PyObject_NEW(PyhlFileCreationProperty,&PyhlFileCreationProperty_Type);
   if(!retv)
      return NULL;
   if(!(retv->props=createHlHdfFileCreationProperty())) {
      setException(PyExc_MemoryError,"Failed to create FileCreationProperty\n");
      _dealloc_pyhlfilecreationproperty(retv);
      retv=NULL;
   }
   return (PyObject*)retv;
}

static PyObject* _pyhl_new_compression(PyObject* self, PyObject* args)
{
   PyhlCompression* retv=NULL;
   HL_CompressionType type;
   if(!PyArg_ParseTuple(args,"i",&type))
      return NULL;
   
   retv=PyObject_NEW(PyhlCompression,&PyhlCompression_Type);
   if(!retv)
      return NULL;
   if(!(retv->compr=newHL_Compression(type))) {
      setException(PyExc_MemoryError,"Failed to create HL_Compression instance\n");
      _dealloc_pyhlcompression(retv);
      retv=NULL;
   }
   return (PyObject*)retv;
}

static PyObject* _pyhl_read_nodelist(PyObject* self, PyObject* args)
{
   HL_NodeList* nodelist=NULL;
   PyhlObject* retv=NULL;
   char* filename=NULL;
   char* frompath=NULL;
   
   if(!PyArg_ParseTuple(args,"s|s",&filename,&frompath))
      return NULL;

   if(!frompath) {
      nodelist = readHL_NodeList(filename);
   } else {
      nodelist = readHL_NodeListFrom(filename,frompath);
   }
	  
   if(!nodelist) {
      char errmsg[256];
      if(!frompath)
	 sprintf(errmsg,"Could not read file '%s'",filename);
      else
	 sprintf(errmsg,"Could not read file '%s' from '%s'",filename,frompath);
      setException(PyExc_IOError,errmsg);
      goto fail;
   }

   if(!(retv=(PyhlObject*)_pyhl_new_nodelist(NULL,NULL))) {
      setException(PyExc_MemoryError,"Could not allocate nodelist instance");
      goto fail;
   }

   /*Change the nodelist*/
   freeHL_NodeList(retv->nodelist);
   retv->nodelist=nodelist;
   return (PyObject*)retv;
 fail:
   freeHL_NodeList(nodelist);
   if(retv)
      _dealloc(retv);
   return NULL;
}

static PyObject* _pyhl_is_file_hdf5(PyObject* self, PyObject* args)
{
   char* filename;
   int val;

   if(!(PyArg_ParseTuple(args,"s",&filename)))
      return NULL;

   val = isHdf5File(filename);

   return PyInt_FromLong((int)val);
}

static PyObject* _pyhl_show_hdf5errors(PyObject* self, PyObject* args)
{
   int yesno;
   if(!PyArg_ParseTuple(args,"i",&yesno))
      return NULL;
   if(yesno==0) {
      HL_disableHdf5ErrorReporting();
   } else {
      HL_enableHdf5ErrorReporting();
   }
   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject* _pyhl_show_hlhdferrors(PyObject* self, PyObject* args)
{
   int yesno;
   if(!PyArg_ParseTuple(args,"i",&yesno))
      return NULL;
   if(yesno==0) {
      HL_setDebugLevel(HLHDF_SILENT);
   } else {
      HL_setDebugLevel(HLHDF_DEBUG);
   }
   Py_INCREF(Py_None);
   return Py_None;
}

/* PyhlObject member methods */
static PyObject* _pyhl_add_node(PyhlObject* self, PyObject* args)
{
   PyObject* inp;
   PyhlNode* pyhlNode;
   HL_Node* aNode;

   if(!PyArg_ParseTuple(args,"O",&inp))
      return NULL;

   if(!PyhlNode_Check(inp)) {
      setException(PyExc_TypeError,"Trying to add node, which not is of PyhlNodeCore type");
      return NULL;
   }

   pyhlNode=(PyhlNode*)inp;
   aNode=pyhlNode->node;

   if(!addNode(self->nodelist,aNode)) {
      setException(PyExc_MemoryError,"Could not add node to nodelist");
      freeHL_Node(aNode);
      return NULL;
   }

   pyhlNode->node=NULL;

   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject* _pyhl_hlwrite(PyhlObject* self, PyObject* args)
{
   char* filename;
   PyObject* props=NULL;
   HL_FileCreationProperty* theProperties=NULL;
   
   if(!PyArg_ParseTuple(args,"s|O",&filename,&props)) {
      return NULL;
   }
   strcpy(self->nodelist->filename,filename);
   if(props!=NULL) {
      if(!PyhlFileCreationProperty_Check(props)) {
	 setException(PyExc_AttributeError,"Second attribute to hlwrite must be of FileCreationProperty");
	 return NULL;
      }
   }

   if(props!=NULL) {
      theProperties = ((PyhlFileCreationProperty*)props)->props;
   }
   if(!writeHL_NodeList(self->nodelist,theProperties,NULL)) {
      setException(PyExc_IOError,"Could not write file");
      return NULL;
   }
   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject* _pyhl_write(PyhlObject* self, PyObject* args)
{
   char* filename;
   PyObject* obj1=NULL;
   PyObject* obj2=NULL;
   
   int doCompress=-1;
   PyObject* props=NULL;
   HL_Compression* theCompression=NULL;
   
   if(!PyArg_ParseTuple(args,"s|OO",&filename,&obj1,&obj2))
      return NULL;

   if(obj1!=NULL) {
      if(PyInt_Check(obj1)) {
	 doCompress=PyInt_AsLong(obj1);
      } else if(PyhlFileCreationProperty_Check(obj1)) {
	 props=obj1;
      } else {
	 setException(PyExc_AttributeError,"write method should be called with write(filename[,zlib compression level(int)][,file creation property])");
	 return NULL;
      }
   }
   if(obj2!=NULL) {
      if(PyInt_Check(obj2)) {
	 if(doCompress!=-1) {
	    setException(PyExc_AttributeError,"it is not meaningful or possible to call write with both second and third argument as integers, format is: "
			 "write(filename[,zlib compression level(int)][,file creation property])");
	    return NULL;
	 }
	 doCompress=PyInt_AsLong(obj2);
      } else if(PyhlFileCreationProperty_Check(obj2)) {
	 if(props!=NULL) {
	    setException(PyExc_AttributeError,"it is not meaningful or possible to call write with both second and third argument as FileCreationProperty instances"
			 ", format is: write(filename[,zlib compression level(int)][,file creation property])");
	    return NULL;
	 }
	 props = obj2;
      } else {
	 setException(PyExc_AttributeError,"write method should be called with write(filename[,zlib compression level(int)][,file creation property])");
	 return NULL;
      }
   }
   
   strcpy(self->nodelist->filename,filename);
   if(doCompress!=-1) {
      theCompression = newHL_Compression(CT_ZLIB);
      theCompression->level = doCompress;
   }
   
   if(!writeHL_NodeList(self->nodelist,(props!=NULL)?((PyhlFileCreationProperty*)props)->props:NULL,theCompression)) {
      setException(PyExc_IOError,"Could not write hdf file");
      if(theCompression) {
	 freeHL_Compression(theCompression);
      }
      return NULL;
   }

   if(theCompression) {
      freeHL_Compression(theCompression);
   }

   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject* _pyhl_update(PyhlObject* self, PyObject* args)
{
   int doCompress=6;
   if(!PyArg_ParseTuple(args,"|i",&doCompress))
      return NULL;   
   if(!updateNodeList(self->nodelist,doCompress)) {
      setException(PyExc_IOError,"Could not update file");
      return NULL;
   }
   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject* _pyhl_get_node_names(PyhlObject* self, PyObject* args)
{
   PyObject* retv=NULL;
   PyObject* pyo=NULL;
   int i;
   char errbuf[256];
   if(!(retv=PyDict_New())) {
      setException(PyExc_MemoryError,"Could not allocate dictionary");
      return NULL;
   }
   for(i=0;i<self->nodelist->nNodes;i++) {
      switch(self->nodelist->nodes[i]->type) {
      case ATTRIBUTE_ID:
	 pyo=PyInt_FromLong(ATTRIBUTE_ID);
	 break;
      case DATASET_ID:
	 pyo=PyInt_FromLong(DATASET_ID);
	 break;
      case TYPE_ID:
	 pyo=PyInt_FromLong(TYPE_ID);
	 break;
      case GROUP_ID:
	 pyo=PyInt_FromLong(GROUP_ID);
	 break;
      case REFERENCE_ID:
	 pyo=PyInt_FromLong(REFERENCE_ID);
	 break;
      default:
	 sprintf(errbuf,"Unknown type for node '%s': '%d'",self->nodelist->nodes[i]->name,
		 self->nodelist->nodes[i]->type);
	 setException(PyExc_TypeError,errbuf);
	 goto fail;
      }
      if(PyDict_SetItemString(retv,self->nodelist->nodes[i]->name,pyo)==-1) {
	 setException(PyExc_AttributeError,"Failed to set dictionary item");
	 goto fail;
      }
      Py_XDECREF(pyo);
      pyo=NULL;
   }

   return retv;
 fail:
   if(pyo) {
      Py_XDECREF(pyo);
   }
   if(retv) {
      Py_XDECREF(retv);
   }
   return NULL;
}

static PyObject* _pyhl_select_all(PyhlObject* self, PyObject* args)
{
   selectAllNodes(self->nodelist);
   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject* _pyhl_select_metadata(PyhlObject* self, PyObject* args)
{
   selectMetadataNodes(self->nodelist);
   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject* _pyhl_select_node(PyhlObject* self, PyObject* args)
{
   char* nodename;
   char errbuf[256];

   if(!PyArg_ParseTuple(args,"s",&nodename))
      return NULL;
   if(!selectNode(self->nodelist,nodename)) {
      sprintf(errbuf,"Could not select node '%s'",nodename);
      setException(PyExc_AttributeError,errbuf);
      goto fail;
   }
   Py_INCREF(Py_None);
   return Py_None;
 fail:
   return NULL;
}

static PyObject* _pyhl_deselect_node(PyhlObject* self, PyObject* args)
{
   char* nodename;
   char errbuf[256];

   if(!PyArg_ParseTuple(args,"s",&nodename))
      return NULL;
   if(!deselectNode(self->nodelist,nodename)) {
      sprintf(errbuf,"Could not select node '%s'",nodename);
      setException(PyExc_AttributeError,errbuf);
      goto fail;
   }
   Py_INCREF(Py_None);
   return Py_None;
 fail:
   return NULL;
}

static PyObject* _pyhl_fetch(PyhlObject* self, PyObject* args)
{
   if(!fetchMarkedNodes(self->nodelist)) {
      setException(PyExc_IOError,"Could not fetch selected nodes");
      goto fail;
   }
   Py_INCREF(Py_None);
   return Py_None;
 fail:
   return NULL;
}

static PyObject* _pyhl_get_node(PyhlObject* self, PyObject* args)
{
   char* nodename;
   char errbuf[256];
   HL_Node* node=NULL;
   PyhlNode* retv=NULL;
   PyObject* myArgs=NULL;

   if(!PyArg_ParseTuple(args,"s",&nodename))
      return NULL;

   if(!(node=getNode(self->nodelist,nodename))) {
      sprintf(errbuf,"Could not get node '%s'",nodename);
      setException(PyExc_IOError,errbuf);
      goto fail;
   }

   if(!(myArgs=Py_BuildValue("(is)",node->type,node->name))) {
      setException(PyExc_AttributeError,"Could not create argument tuple for allocating node");
      goto fail;
   }

   if(!(retv=(PyhlNode*)_pyhl_new_node((PyObject*)self,myArgs))) {
      setException(PyExc_AttributeError,"Could not create node to return");
      goto fail;
   }

   freeHL_Node(retv->node);

   retv->node=copyHL_Node(node);
   
   Py_XDECREF(myArgs);
   return (PyObject*)retv;
 fail:
   Py_XDECREF(myArgs);
   if(retv)
      _dealloc_pyhlnode(retv);
   return NULL;
}

/* PyhlNode member methods */
static PyObject* _pyhl_node_set_scalar_value(PyhlNode* self, PyObject* args)
{
   int itemSize;
   char* hltypename;
   PyObject* data;
   hid_t lhid=-1;
   hid_t tmpHid=-1;
   int chkVal=0;

   if(!PyArg_ParseTuple(args,"iOsl",&itemSize,&data,&hltypename,&lhid))
      return NULL;

   if(!self->node) {
      setException(PyExc_AttributeError,"The responsibility of the node has been dropped, probably by doing a addNode");
      return NULL;
   }

   if(strcmp(hltypename,"string")!=0 &&
      strcmp(hltypename,"compound")!=0 &&
      (tmpHid=translateCharToDatatype(hltypename))==-1) {
      char errmsg[256];
      sprintf(errmsg,"No datatype called '%s'.",hltypename);
      setException(PyExc_AttributeError,errmsg);
      return NULL;
   }

   if(tmpHid!=-1) {
      /* Well, ordinary attribute, just make sure the space is enough */
      size_t typeSize=H5Tget_size(tmpHid);
      chkVal=0;
      switch(H5Tget_class(tmpHid)) {
      case H5T_INTEGER: {
	 if(typeSize<=sizeof(char)) {
	    char mData = (char)PyInt_AsLong(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,hltypename,-1);
	 } else if(typeSize<=sizeof(short)) {
	    short mData = (char)PyInt_AsLong(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,hltypename,-1);
	 } else if(typeSize<=sizeof(int)) {
	    int mData = (int)PyInt_AsLong(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,hltypename,-1);
	 } else if(typeSize<=sizeof(long)) {
	    long mData = (long)PyInt_AsLong(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,hltypename,-1);
	 } else {
	    long long mData = (long long)PyLong_AsLongLong(data);
	    if(PyErr_Occurred()) {
	       setException(PyExc_AttributeError,"Could not translate to long long, probably not supported");
	       goto fail;
	    }
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,hltypename,-1);
	 }
	 break;
      }
      case H5T_FLOAT: {
	 if(typeSize<=sizeof(float)) {
	    float mData = (float)PyFloat_AsDouble(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,hltypename,-1);
	 } else if (typeSize<=sizeof(double)) {
	    double mData = (double)PyFloat_AsDouble(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,hltypename,-1);
	 } else {
	    fprintf(stderr,"Whoaa, greater float than double not supported\n");
	 }
	 break;
      }
      default: {
	 char errmsg[256];
	 sprintf(errmsg,"_pyhl_node_set_scalar_value Could not handle type in switch,"
		 "type=%d\n", (int)H5Tget_class(tmpHid));
	 setException(PyExc_TypeError,errmsg);
	 goto fail;
      }
      }
      H5Tclose(tmpHid);
      tmpHid=-1;
      if(!chkVal) {
	 setException(PyExc_TypeError,"Could not set scalar value");
	 goto fail;
      }
   } else if(strcmp(hltypename,"string")==0) {
      char* mData = PyString_AsString(data);
      if(!mData || !setScalarValue(self->node,strlen(mData)+1,(unsigned char*)mData,hltypename,-1)) {
	 setException(PyExc_ValueError,"Could not set scalar string");
	 goto fail;
      }
   } else if(strcmp(hltypename,"compound")==0) {
      char* mData = PyString_AsString(data);
      if(!mData || !setScalarValue(self->node,itemSize,(unsigned char*)mData,hltypename,lhid)) {
	 setException(PyExc_ValueError,"Could not set scalar compound value");
	 goto fail;
      }
   }

   if(tmpHid>=0)
      H5Tclose(tmpHid);
   Py_INCREF(Py_None);
   return Py_None;
 fail:
   if(tmpHid>=0)
      H5Tclose(tmpHid);

   return NULL;
}

static PyObject* _pyhl_node_set_array_value(PyhlNode* self, PyObject* args)
{
   PyObject* pydims;
   PyObject* data;
   hid_t lhid=-1;
   char* hltypename;
   int itemSize;
   int n;
   hsize_t dims[4];
   int i;
   int ndim;
   PyObject* pyo=NULL;
   size_t tmpSize;

   if(!PyArg_ParseTuple(args,"iOOsl",&itemSize,&pydims,&data,&hltypename,&lhid))
      return NULL;

   if(!self->node) {
      setException(PyExc_AttributeError,"The responsibility of the node has been dropped, probably by doing a addNode");
      return NULL;
   }

   if(PyObject_Length(pydims)>4) {
      setException(PyExc_ValueError,"Not more than 4 dimensions are allowed");
      goto fail;
   }

   if(PyArray_Check(data) && isFormatSupported(hltypename)) {
      /*Ahh plain array*/
      if(PyObject_Length(pydims)!=((PyArrayObject*)data)->nd) {
	 setException(PyExc_ValueError,"Rank of Array != no of dims");
	 goto fail;
      }
      ndim=PyObject_Length(pydims);
      for(i=0;i<ndim;i++) { /*Extract dimensions*/
	 if(!(pyo=PySequence_GetItem(pydims,i))) {
	    setException(PyExc_AttributeError,"Could not get list item");
	    goto fail;
	 }
	 dims[i] = PyInt_AsLong(pyo);
	 Py_XDECREF(pyo);
	 pyo=NULL;
	 if(dims[i]!=((PyArrayObject*)data)->dimensions[i]) {
	    setException(PyExc_AttributeError,"Array dimensions != list dimensions");
	    goto fail;
	 }
      }
      if((tmpSize = whatSizeIsHdfFormat(hltypename))<0) {
	 setException(PyExc_ValueError,"Could not determine size");
	 goto fail;
      }
      if(tmpSize!=((PyArrayObject*)data)->descr->elsize) {
	 setException(PyExc_ValueError,"Type sizes are different between format and array");
	 goto fail;
      }
      if(!setArrayValue(self->node,tmpSize,ndim,dims,(unsigned char*)((PyArrayObject*)data)->data,
			hltypename,-1)) {
	 setException(PyExc_AttributeError,"Could not set array data");
	 goto fail;
      }
   } else if(PyList_Check(data) && (strcmp(hltypename,"string")==0 || isFormatSupported(hltypename))) {
      /* Okie, list of something */
      int maxstrlen=0;
      char* tmpData=NULL;
      hid_t strtype=-1;

      if(PyObject_Length(pydims)!=1) {
	 setException(PyExc_ValueError,"When using lists, the only allowed rank is 1.");
	 goto fail;
      }
      n=PyObject_Length(data);
      if(strcmp(hltypename,"string")==0) {
	 /*Calculate the longest string*/
	 for(i=0;i<n;i++) {
	    char* tmpstr;
	    if(!(pyo=PySequence_GetItem(data,i))) {
	       setException(PyExc_AttributeError,"Could not get list item");
	       goto fail;
	    }
	    if(!(tmpstr = PyString_AsString(pyo))) {
	       setException(PyExc_AttributeError,"Could not translate data to a string");
	       goto fail;
	    }
	    if(strlen(tmpstr)>maxstrlen)
	       maxstrlen=strlen(tmpstr);
	    Py_XDECREF(pyo);
	    pyo=NULL;
	 }
	 itemSize = maxstrlen + 1;
	 if(!(tmpData=malloc(itemSize * n))) {
	    setException(PyExc_MemoryError,"Could not allocate memory for strings");
	    goto fail;
	 }
	 memset(tmpData,0,itemSize*n);
	 for(i=0;i<n;i++) {
	    char* tmpstr;
	    pyo = PySequence_GetItem(data,i);
	    tmpstr = PyString_AsString(pyo);
	    strncpy(&tmpData[i*itemSize],tmpstr,strlen(tmpstr));
	 }
	 if((strtype = createStringType(itemSize))<0) {
	    setException(PyExc_MemoryError,"Could not create a string type");
	    free(tmpData);
	    goto fail;
	 }
	 dims[0]=n;
	 if(!(setArrayValue(self->node,itemSize,1,dims,(unsigned char*)tmpData,hltypename,strtype))) {
	    setException(PyExc_AttributeError,"Could not set array data");
	    H5Tclose(strtype);
	    free(tmpData);
	    goto fail;
	 }
      } else {
	 double* doubleptr=NULL;
	 float* floatptr=NULL;
	 int* intptr=NULL;
	 long* longptr=NULL;
	 if((tmpSize = whatSizeIsHdfFormat(hltypename))<0) {
	    setException(PyExc_ValueError,"Could not determine size");
	    goto fail;
	 }
	 if(strcmp(hltypename,"double")==0) {
	    doubleptr = malloc(sizeof(double)*n);
	    tmpData = (char*)doubleptr;
	 } else if(strcmp(hltypename,"float")==0) {
	    floatptr = malloc(sizeof(float)*n);
	    tmpData = (char*)floatptr;
	 } else if(strcmp(hltypename,"int")==0) {
	    intptr = malloc(sizeof(int)*n);
	    tmpData = (char*)intptr;
	 } else if(strcmp(hltypename,"long")==0) {
	    longptr = malloc(sizeof(long)*n);
	    tmpData = (char*)longptr;
	 } else {
	    setException(PyExc_TypeError,"Only supported types on list are double,float,int and long");
	    goto fail;
	 }
	 if(!tmpData) {
	    setException(PyExc_MemoryError,"Could not allocate memory for datatype");
	    goto fail;
	 }

	 for(i=0;i<n;i++) {
	    if(!(pyo=PySequence_GetItem(data,i))) {
	       setException(PyExc_AttributeError,"Could not get list item");
	       free(tmpData);
	       goto fail;
	    }
	    if(doubleptr) {
	       doubleptr[i] = (double)PyFloat_AsDouble(pyo);
	    } else if(floatptr) {
	       floatptr[i] = (float)PyFloat_AsDouble(pyo);
	    } else if(intptr) {
	       intptr[i] = (int)PyInt_AsLong(pyo);
	    } else if(longptr) {
	       longptr[i] = (long)PyInt_AsLong(pyo);
	    } else {
	       setException(PyExc_AttributeError,"Whoa, something very strange happened");
	       free(tmpData);
	       goto fail;
	    }
	    Py_XDECREF(pyo);
	    pyo=NULL;
	 }
	 dims[0]=n;
	 if(!(setArrayValue(self->node,tmpSize,1,dims,(unsigned char*)tmpData,hltypename,-1))) {
	    setException(PyExc_AttributeError,"Could not set array data");
	    free(tmpData);
	    goto fail;
	 }
      }
      if(tmpData) free(tmpData);
   } else if(strcmp(hltypename,"compound")==0) {
      size_t tmpLen=0;
      char* tmpData=PyString_AsString(data);
      if(!tmpData) {
	 setException(PyExc_AttributeError,"Could not translate data to a string");
	 goto fail;
      }
      ndim=PyObject_Length(pydims);
      for(i=0;i<ndim;i++) {
	 if(!(pyo=PySequence_GetItem(pydims,i))) {
	    setException(PyExc_AttributeError,"Could not get list item");
	    goto fail;
	 }
	 dims[i] = PyInt_AsLong(pyo);
	 Py_XDECREF(pyo);
	 pyo=NULL;
      }
      tmpLen=1;
      for(i=0;i<ndim;i++) {
	 tmpLen*=dims[i];
      }
      tmpLen*=itemSize;
      if(PyObject_Length(data)!=tmpLen) {
	 setException(PyExc_AttributeError,"There is some inconsistency between data size and calculated datasize");
	 goto fail;
      }
      if(!setArrayValue(self->node,itemSize,ndim,dims,(unsigned char*)tmpData,hltypename,lhid)) {
	 setException(PyExc_AttributeError,"Could not set array data");
	 goto fail;
      }
   } else {
      setException(PyExc_AttributeError,"Only support arrays and compound types right now");
      goto fail;
   }

   if(pyo) {
      Py_XDECREF(pyo);
   }
   Py_INCREF(Py_None);
   return Py_None;
 fail:
   if(pyo) {
      Py_XDECREF(pyo);
   }
   return NULL;
}

static PyObject* _pyhl_node_commit(PyhlNode* self, PyObject* args)
{
   hid_t dType;
   if(!PyArg_ParseTuple(args,"l",&dType))
      return NULL;
   if(!self->node->type==TYPE_ID) {
      setException(PyExc_AttributeError,"Trying to commit a node which not is a type node");
      return NULL;
   }
   commitDatatype(self->node,dType);
   Py_INCREF(Py_None);
   return Py_None;
}
   
static PyObject* _pyhl_node_name(PyhlNode* self, PyObject* args)
{
   return PyString_FromString(self->node->name);
}

static PyObject* _pyhl_node_type(PyhlNode* self, PyObject* args)
{
   return PyInt_FromLong(self->node->type);
}

static PyObject* _pyhl_node_dims(PyhlNode* self,PyObject* args)
{
   PyObject* retv=NULL;
   int i;
   PyObject* pyo=NULL;
   if(!(retv=PyList_New(0))) {
      return NULL;
   }
   for(i=0;i<self->node->ndims;i++) {
      if(!(pyo = PyInt_FromLong(self->node->dims[i]))) {
	 setException(PyExc_ValueError,"Could not create py integer");
	 goto fail;
      }
      if(PyList_Append(retv,pyo)==-1) {
	 setException(PyExc_ValueError,"Could not append list item");
	 goto fail;
      }
      Py_XDECREF(pyo);
      pyo=NULL;
   }
   Py_XDECREF(pyo);
   return retv;
 fail:
   Py_XDECREF(pyo);
   Py_XDECREF(retv);
   return NULL;
}

static PyObject* _pyhl_node_format(PyhlNode* self,PyObject* args)
{
   return PyString_FromString(self->node->format);
}

static PyObject* _pyhl_node_data(PyhlNode* self, PyObject* args)
{
   PyObject* retv=NULL;
   hid_t tmpHid=-1;
   size_t typeSize;
   char errbuf[256];
   int* dims=NULL;
   int i;
   size_t npts;

   if(self->node->typeId>=0) {
      tmpHid=H5Tcopy(self->node->typeId);
   } else if(isFormatSupported(self->node->format)) {
      tmpHid=translateCharToDatatype(self->node->format);
   } else {
      setException(PyExc_AttributeError,"Strange type, can't handle");
      return NULL;
   }

   typeSize = H5Tget_size(tmpHid);

   if(self->node->ndims==0) { /*Scalar*/
      switch(H5Tget_class(tmpHid)) {
      case H5T_INTEGER: {
	 if(typeSize<=sizeof(char)) {
	    char v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else if(typeSize<=sizeof(short)) {
	    short v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else if(typeSize<=sizeof(int)) {
	    int v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else if(typeSize<=sizeof(long)) {
	    long v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else {
	    long long v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyLong_FromLongLong(v);
	 }
	 break;
      }
      case H5T_FLOAT: {
	 if(typeSize<=sizeof(float)) {
	    float v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyFloat_FromDouble((double)v);
	 } else if(typeSize<=sizeof(double)) {
	    double v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyFloat_FromDouble(v);
	 } else {
	    fprintf(stderr,"Whoaa, greater float than double not supported\n");
	 }
	 break;
      }
      case H5T_COMPOUND: {
	 retv=PyString_FromStringAndSize((char*)self->node->data,typeSize);
	 break;
      }
      case H5T_STRING: {
	 retv=PyString_FromStringAndSize((char*)self->node->data,typeSize-1);
	 break;
      }
      default: {
	 sprintf(errbuf,"Unrecognized class %d",(int)H5Tget_class(tmpHid));
	 setException(PyExc_TypeError,errbuf);
	 break;
      }
      }
   } else { /*Simple*/
      if(!(dims = malloc(sizeof(int)*self->node->ndims))) {
	 setException(PyExc_MemoryError,"Could not allocate dims");
	 goto fail;
      }
      switch(H5Tget_class(tmpHid)) {
      case H5T_INTEGER:
      case H5T_FLOAT: {
	 int iformat=pyarraytypeFromHdfType(self->node->format);
	 int nbytes;
	 if(iformat==-1) {
	    sprintf(errbuf,"Unrecognized datatype %s",self->node->format);
	    setException(PyExc_TypeError,errbuf);
	    goto fail;
	 }
	 for(i=0;i<self->node->ndims;i++)
	    dims[i]=(int)self->node->dims[i];

	 /* Replaced PyArray_FromDimsAndData with this */
	 retv = PyArray_FromDims(self->node->ndims,dims,iformat);
	 if(!retv) {
	    setException(PyExc_MemoryError,"Could not create array");
	    goto fail;
	 }
	 nbytes = 1;
	 for(i=0;i<self->node->ndims;i++)
	    nbytes *= dims[i];
	 nbytes *= ((PyArrayObject*)retv)->descr->elsize;

	 memcpy(((PyArrayObject*)retv)->data,(unsigned char*)self->node->data,nbytes);
	 break;
      }
      case H5T_COMPOUND: {
	 npts=1;
	 for(i=0;i<self->node->ndims;i++)
	    npts*=self->node->dims[i];
	 npts*=typeSize;
	 retv=PyString_FromStringAndSize((char*)self->node->data,npts);
	 break;
      }
      case H5T_STRING: {
	 if(self->node->ndims!=1) {
	    /* Dunno how to represent a multi-dim array of strings */
	    npts=1;
	    for(i=0;i<self->node->ndims;i++)
	       npts*=self->node->dims[i];
	    npts*=typeSize;
	    retv=PyString_FromStringAndSize((char*)self->node->data,npts-1);
	 } else {
	    retv = PyList_New(0);
	    for(i=0;retv && i<self->node->dims[0];i++) {
	       PyObject* pyo = PyString_FromStringAndSize((char*)(&self->node->data[i*typeSize]),typeSize-1);
	       if(!pyo) {
		  setException(PyExc_MemoryError,"Could not create string");
		  goto fail;
	       }
	       if(PyList_Append(retv,pyo)==-1) {
		  Py_XDECREF(pyo);
		  goto fail;
	       }
	       Py_XDECREF(pyo);
	    }
	 }
	 break;
      }
      default: {
	 sprintf(errbuf,"Unrecognized class %d",(int)H5Tget_class(tmpHid));
	 setException(PyExc_TypeError,errbuf);
	 break;
      }
      }
   }
   if(dims)
      free(dims);
   if(tmpHid>=0)
      H5Tclose(tmpHid);
   return retv;
 fail:
   Py_XDECREF(retv);
   if(dims)
      free(dims);
   if(tmpHid>=0)
      H5Tclose(tmpHid);
   return NULL;
}

static PyObject* _pyhl_node_rawdata(PyhlNode* self, PyObject* args)
{
   PyObject* retv=NULL;
   hid_t tmpHid=-1;
   size_t typeSize;
   char errbuf[256];
   int* dims=NULL;
   int i;
   size_t npts;

   if(self->node->rawdata==NULL) {
      setException(PyExc_AttributeError,"Rawdata has not been read for this node");
      return NULL;
   }
      
   if(self->node->typeId>=0) {
      tmpHid=H5Tcopy(self->node->typeId);
   } else if(isFormatSupported(self->node->format)) {
      tmpHid=translateCharToDatatype(self->node->format);
   } else {
      setException(PyExc_AttributeError,"Strange type, can't handle");
      return NULL;
   }

   typeSize = self->node->dSize; /* Using raw type */

   if(self->node->ndims==0) { /*Scalar*/
      switch(H5Tget_class(tmpHid)) {
      case H5T_INTEGER: {
	 if(typeSize<=sizeof(char)) {
	    char v;
	    memcpy(&v,self->node->rawdata,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else if(typeSize<=sizeof(short)) {
	    short v;
	    memcpy(&v,self->node->rawdata,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else if(typeSize<=sizeof(int)) {
	    int v;
	    memcpy(&v,self->node->rawdata,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else if(typeSize<=sizeof(long)) {
	    long v;
	    memcpy(&v,self->node->rawdata,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else {
	    long long v;
	    memcpy(&v,self->node->rawdata,typeSize);
	    retv=PyLong_FromLongLong(v);
	 }
	 break;
      }
      case H5T_FLOAT: {
	 if(typeSize<=sizeof(float)) {
	    float v;
	    memcpy(&v,self->node->rawdata,typeSize);
	    retv=PyFloat_FromDouble((double)v);
	 } else if(typeSize<=sizeof(double)) {
	    double v;
	    memcpy(&v,self->node->rawdata,typeSize);
	    retv=PyFloat_FromDouble(v);
	 } else {
	    fprintf(stderr,"Whoaa, greater float than double not supported\n");
	 }
	 break;
      }
      case H5T_COMPOUND: {
	 retv=PyString_FromStringAndSize((char*)self->node->rawdata,typeSize);
	 break;
      }
      case H5T_STRING: {
	 retv=PyString_FromStringAndSize((char*)self->node->rawdata,typeSize-1);
	 break;
      }
      default: {
	 sprintf(errbuf,"Unrecognized class %d",(int)H5Tget_class(tmpHid));
	 setException(PyExc_TypeError,errbuf);
	 break;
      }
      }
   } else { /*Simple*/
      if(!(dims = malloc(sizeof(int)*self->node->ndims))) {
	 setException(PyExc_MemoryError,"Could not allocate dims");
	 goto fail;
      }
      switch(H5Tget_class(tmpHid)) {
      case H5T_INTEGER:
      case H5T_FLOAT: {
	 int iformat=pyarraytypeFromHdfType(self->node->format);
	 int nbytes;
	 if(iformat==-1) {
	    sprintf(errbuf,"Unrecognized datatype %s",self->node->format);
	    setException(PyExc_TypeError,errbuf);
	    goto fail;
	 }
	 for(i=0;i<self->node->ndims;i++)
	    dims[i]=(int)self->node->dims[i];

	 /* Replaced PyArray_FromDimsAndData with this */
	 retv = PyArray_FromDims(self->node->ndims,dims,iformat);
	 if(!retv) {
	    setException(PyExc_MemoryError,"Could not create array");
	    goto fail;
	 }
	 nbytes = 1;
	 for(i=0;i<self->node->ndims;i++)
	    nbytes *= dims[i];
	 nbytes *= ((PyArrayObject*)retv)->descr->elsize;
	 memcpy(((PyArrayObject*)retv)->data,(unsigned char*)self->node->rawdata,nbytes);
	 break;
      }
      case H5T_COMPOUND: {
	 npts=1;
	 for(i=0;i<self->node->ndims;i++)
	    npts*=self->node->dims[i];
	 npts*=typeSize;
	 retv=PyString_FromStringAndSize((char*)self->node->rawdata,npts);
	 break;
      }
      case H5T_STRING: {
	 if(self->node->ndims!=1) {
	    /* Dunno how to represent a multi-dim array of strings */
	    npts=1;
	    for(i=0;i<self->node->ndims;i++)
	       npts*=self->node->dims[i];
	    npts*=typeSize;
	    retv=PyString_FromStringAndSize((char*)self->node->rawdata,npts-1);
	 } else {
	    retv = PyList_New(0);
	    for(i=0;retv && i<self->node->dims[0];i++) {
	       PyObject* pyo = PyString_FromStringAndSize((char*)(&self->node->rawdata[i*typeSize]),typeSize-1);
	       if(!pyo) {
		  setException(PyExc_MemoryError,"Could not create string");
		  goto fail;
	       }
	       if(PyList_Append(retv,pyo)==-1) {
		  Py_XDECREF(pyo);
		  goto fail;
	       }
	       Py_XDECREF(pyo);
	    }
	 }
	 break;
      }
      default: {
	 sprintf(errbuf,"Unrecognized class %d",(int)H5Tget_class(tmpHid));
	 setException(PyExc_TypeError,errbuf);
	 break;
      }
      }
   }
   if(dims)
      free(dims);
   if(tmpHid>=0)
      H5Tclose(tmpHid);
   return retv;
 fail:
   Py_XDECREF(retv);
   if(dims)
      free(dims);
   if(tmpHid>=0)
      H5Tclose(tmpHid);
   return NULL;
}

static PyObject* getPythonObjectFromNode(HL_CompoundTypeAttribute* descr ,unsigned char* data, int idx)
{
   PyObject* pyo = NULL;

   int offset = descr->offset + idx * descr->size;
   int debug = 0;
   
   char tmpstr[300];
   memset(tmpstr,0,300);

   if(debug) {
      printf("name = %s, offset = %d, descr->offset = %d, idx = %d, descr->size=%d, ",descr->attrname, offset, descr->offset,
	     idx, descr->size);
   }

   if(strcmp(descr->format,"string")==0) {
      int len = strlen((char*)&data[offset]);
      char* dummystr = "";
      if( len > 0 && len <=descr->size) {
	 if(debug)
	    strncpy(&tmpstr[0],(char*)&data[offset],len);
	 pyo = PyString_FromStringAndSize((char*)&data[offset],len);
      } else if(len <= 0) {
	 if(debug)
	    strncpy(&tmpstr[0],dummystr,len);
	 pyo = PyString_FromString(dummystr);
      } else if(len >= descr->size) {
	 if(debug)
	    strncpy(&tmpstr[0],(char*)&data[offset],len);
	 pyo = PyString_FromStringAndSize((char*)&data[offset],descr->size);
      }
      if(debug)
	 printf("val = '%s'\n",tmpstr);
   } else if(strcmp(descr->format,"schar")==0 ||
	     strcmp(descr->format,"uchar")==0 ||
	     strcmp(descr->format,"char")==0) {
      char val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%c\n",val);
      pyo = PyString_FromStringAndSize(&val,1);
   } else if(strcmp(descr->format,"short")==0 ||
	     strcmp(descr->format,"ushort")==0) {
      short val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%d\n",val);
      pyo = PyInt_FromLong((long)val);
   } else if(strcmp(descr->format,"int")==0 ||
	     strcmp(descr->format,"uint")==0) {
      int val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%d\n",val);
      pyo = PyInt_FromLong((long)val);
   } else if(strcmp(descr->format,"long")==0 ||
	     strcmp(descr->format,"ulong")==0) {
      long val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%d\n",(int)val);
      pyo = PyInt_FromLong((long)val);
   } else if(strcmp(descr->format,"llong")==0) {
      long long val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%d\n",(int)val);
      pyo = PyInt_FromLong((long)val);
   } else if(strcmp(descr->format,"float")==0) {
      float val = 0.0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%f\n",val); /*offset = descr->offset + idx * descr->size;*/
      pyo = PyFloat_FromDouble((double)val);
   } else if(strcmp(descr->format,"double")==0) {
      double val = 0.0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("dval=%f\n",val);
      pyo = PyFloat_FromDouble((double)val);
   } else if(strcmp(descr->format,"ldouble")==0) {
      double val = 0.0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%f\n",val);
      pyo = PyFloat_FromDouble((double)val);
   } else if(strcmp(descr->format,"hsize")==0) {
      hsize_t val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%d\n",(int)val);
      pyo = PyInt_FromLong((long)val);
   } else if(strcmp(descr->format,"hssize")==0) {
      hssize_t val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%d\n",(int)val);
      pyo = PyInt_FromLong((long)val);
   } else if(strcmp(descr->format,"herr")==0) {
      herr_t val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%d\n",val);
      pyo = PyInt_FromLong((long)val);
   } else if(strcmp(descr->format,"hbool")==0) {
      hbool_t val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%d\n",val);
      pyo = PyInt_FromLong((long)val);
   } else {
      char errmsg[256];
      sprintf(errmsg,"Unsupported data type '%s'\n",descr->format);
      setException(PyExc_AttributeError,errmsg);
   }
   return pyo;
}

static PyObject* _pyhl_node_get_compound_data(PyhlNode* self, PyObject* args)
{
   hid_t tmpHid=-1;
   PyObject* retv=NULL;
   PyObject *pyo=NULL,*pyo2=NULL;
   int i,j;
   unsigned char* data;
   HL_CompoundTypeDescription* descr;

   if(self->node->typeId>=0) {
      tmpHid=H5Tcopy(self->node->typeId);
   } else if(isFormatSupported(self->node->format)) {
      tmpHid=translateCharToDatatype(self->node->format);
   } else {
      setException(PyExc_AttributeError,"Strange type, can't handle");
      return NULL;
   }

   if(H5Tget_class(tmpHid)!=H5T_COMPOUND) {
      setException(PyExc_AttributeError,"Only supposed to be used with composite nodes");
      goto fail;
   }
   
   if(self->node->ndims==0 || (self->node->ndims==1 && self->node->dims[0]==1)) { /*Scalar*/
      if(!self->node->compoundDescription) {
	 setException(PyExc_AttributeError,"Node does not have a compound description");
	 goto fail;
      }
      if(!(retv = PyDict_New())) {
	 setException(PyExc_MemoryError,"Could not allocate dicionary\n");
	 goto fail;
      }
      data = self->node->data;
      descr = self->node->compoundDescription;
      for(i=0;i<descr->nAttrs;i++) {
	 pyo = NULL;
	 if(descr->attrs[i]->ndims == 0 || (descr->attrs[i]->ndims == 1 &&
					    descr->attrs[i]->dims[0] == 1)) {
	    if(!(pyo = getPythonObjectFromNode(descr->attrs[i],data,0))) {
	       goto fail;
	    }
	 } else {
	    int ndim=1;
	    if(!(pyo = PyList_New(0))) {
	       goto fail;
	    }
	    for(j=0;j<descr->attrs[i]->ndims;j++) {
	       ndim *= descr->attrs[i]->dims[j];
	    }
	    for(j=0;j<ndim;j++) {
	       if(!(pyo2 = getPythonObjectFromNode(descr->attrs[i],data,j))) {
		  goto fail;
	       }
	       PyList_Append(pyo,pyo2);
	       Py_XDECREF(pyo2);
	       pyo2 = NULL;
	    }
	 }

	 if(PyDict_SetItemString(retv,descr->attrs[i]->attrname,pyo)!=0) {
	    goto fail;
	 }
	 Py_XDECREF(pyo);
	 pyo = NULL;
      }
   } else {
      setException(PyExc_AttributeError,"Method does not support arrays of compound objects yet.");
      goto fail;
   }
   Py_XDECREF(pyo);
   Py_XDECREF(pyo2);
   if(tmpHid>=0)
      H5Tclose(tmpHid);   
   return retv;
fail:
   Py_XDECREF(retv);
   Py_XDECREF(pyo);
   Py_XDECREF(pyo2);
   if(tmpHid>=0)
      H5Tclose(tmpHid);
   return NULL;
}

static struct PyMethodDef methods[] =
{
  {"addNode",(PyCFunction)_pyhl_add_node,1},
  {"write",(PyCFunction)_pyhl_write,1},
  {"update",(PyCFunction)_pyhl_update,1},
  {"getNodeNames",(PyCFunction)_pyhl_get_node_names,1},
  {"selectAll",(PyCFunction)_pyhl_select_all,1},
  {"selectMetadata",(PyCFunction)_pyhl_select_metadata,1},
  {"selectNode",(PyCFunction)_pyhl_select_node,1},
  {"deselectNode",(PyCFunction)_pyhl_deselect_node,1},
  {"fetch",(PyCFunction)_pyhl_fetch,1},
  {"getNode",(PyCFunction)_pyhl_get_node,1},
  {NULL, NULL} /* sentinel */
};

static struct PyMethodDef node_methods[] =
{
   {"setScalarValue",(PyCFunction)_pyhl_node_set_scalar_value,1},
   {"setArrayValue",(PyCFunction)_pyhl_node_set_array_value,1},
   {"commit",(PyCFunction)_pyhl_node_commit,1},
   
   /* Inquiry options */
   {"name",(PyCFunction)_pyhl_node_name,1},
   {"type",(PyCFunction)_pyhl_node_type,1},
   {"dims",(PyCFunction)_pyhl_node_dims,1},
   {"format",(PyCFunction)_pyhl_node_format,1},
   {"data",(PyCFunction)_pyhl_node_data,1},
   {"rawdata",(PyCFunction)_pyhl_node_rawdata,1},
   {"compound_data",(PyCFunction)_pyhl_node_get_compound_data,1},
   {NULL, NULL} /* sentinel */
};

static struct PyMethodDef filecreationproperty_methods[] =
{
   {"version",NULL},
   {"userblock",NULL},
   {"sizes",NULL},
   {"sym_k",NULL},
   {"istore_k",NULL},
   {"meta_block_size",NULL},
   {NULL,NULL}   
};

static struct PyMethodDef compression_methods[] =
{
   {"type",NULL},
   {"level",NULL},
   {"szlib_mask",NULL},
   {"szlib_px_per_block",NULL},
   {"H5_SZIP_CHIP_OPTION_MASK",NULL},
   {"H5_SZIP_ALLOW_K13_OPTION_MASK",NULL},
   {"H5_SZIP_EC_OPTION_MASK",NULL},
   {"H5_SZIP_NN_OPTION_MASK",NULL},
   {NULL,NULL}
};

static PyObject* _getattr(PyhlObject* self, char* name)
{
    PyObject* res;

    if(strcmp(name,"ATTRIBUTE_ID")==0)
       return PyInt_FromLong(ATTRIBUTE_ID);
    if(strcmp(name,"GROUP_ID")==0)
       return PyInt_FromLong(GROUP_ID);
    if(strcmp(name,"DATASET_ID")==0)
       return PyInt_FromLong(DATASET_ID);
    if(strcmp(name,"TYPE_ID")==0)
       return PyInt_FromLong(TYPE_ID);
    if(strcmp(name,"REFERENCE_ID")==0)
       return PyInt_FromLong(REFERENCE_ID);
    
    res = Py_FindMethod(methods, (PyObject*) self, name);
    if (res)
	return res;

    PyErr_Clear();

    PyErr_SetString(PyExc_AttributeError, name);

    return NULL;
}

static PyObject* _getattr_node(PyhlNode* self, char* name)
{
    PyObject* res;

    res = Py_FindMethod(node_methods, (PyObject*) self, name);
    if (res)
	return res;

    PyErr_Clear();

    PyErr_SetString(PyExc_AttributeError, name);

    return NULL;
}

static PyObject* _getattr_filecreationproperty(PyhlFileCreationProperty* self, char* name)
{
    PyObject* res;

   if(strcmp(name,"version")==0) {
      return Py_BuildValue("(iiii)",self->props->version.super,self->props->version.freelist,
			   self->props->version.stab,self->props->version.shhdr);
   } else if(strcmp(name,"userblock")==0) {
      return PyInt_FromLong(self->props->userblock);
   } else if(strcmp(name,"sizes")==0) {
      return Py_BuildValue("(ii)",self->props->sizes.sizeof_addr,self->props->sizes.sizeof_size);
   } else if(strcmp(name,"sym_k")==0) {
      return Py_BuildValue("(ii)",self->props->sym_k.ik,self->props->sym_k.lk);
   } else if(strcmp(name,"istore_k")==0) {
      return PyInt_FromLong(self->props->istore_k);
   } else if(strcmp(name,"meta_block_size")==0) {
      return PyInt_FromLong(self->props->meta_block_size);
   }
   
   res = Py_FindMethod(filecreationproperty_methods, (PyObject*) self, name);
   if (res)
      return res;
   
   PyErr_Clear();
   
   PyErr_SetString(PyExc_AttributeError, name);
   
   return NULL;   
}	

static int _setattr_filecreationproperty(PyhlFileCreationProperty* self, char* name, PyObject* val)
{
   char errmsg[256];
   if(strcmp(name,"userblock")==0) {
      hsize_t tmpv;
      hsize_t tmpv2;
      hsize_t mask=1;
      int found_setbit=0;
      int number_of_shifts=sizeof(hsize_t)*8;
      int i;
      tmpv = PyInt_AsLong(val);
      if(tmpv<512 && tmpv!=0) {
	 setException(PyExc_AttributeError,"The userblock size should be set to 0 or any power of 2 greater or equal to 512\n");
	 return -1;
      }
      tmpv2=tmpv;
      for(i=0;i<number_of_shifts;i++) {
	 if(tmpv2&mask) {
	    if(found_setbit) {
	       setException(PyExc_AttributeError,"The userblock size should be set to 0 or any power of 2 greater or equal to 512\n");
	       return -1;
	    }
	    found_setbit=1;
	 }
	 tmpv2=tmpv2>>1;
      }
      self->props->userblock=tmpv;
      return 0;
   } else if(strcmp(name,"sizes")==0) {
      size_t sizeof_addr,sizeof_size;
      if(!PyArg_ParseTuple(val,"ii",&sizeof_addr,&sizeof_size)) {
	 setException(PyExc_AttributeError,"sizes should be set with a tuble (sizeof_addr,sizeof_size)\n");
	 return -1;
      }
      if(sizeof_addr != 0 && sizeof_addr != 2 && sizeof_addr != 4 && sizeof_addr != 8 && sizeof_addr != 16) {
	 setException(PyExc_AttributeError,"first argument in sizes (sizeof_addr) must be one of 0,2,4,8 or 16\n");
	 return -1;
      }
      if(sizeof_size != 0 && sizeof_size != 2 && sizeof_size != 4 && sizeof_size != 8 && sizeof_size != 16) {
	 setException(PyExc_AttributeError,"second argument in sizes (sizeof_size) must be one of 0,2,4,8 or 16\n");
	 return -1;
      }
      if(sizeof_size!=0) {
	 self->props->sizes.sizeof_size = sizeof_size;
      }
      if(sizeof_addr!=0) {
	 self->props->sizes.sizeof_addr = sizeof_addr;
      }
      return 0;
   } else if(strcmp(name,"sym_k")==0) {
      int ik,lk;
      if(!PyArg_ParseTuple(val,"ii",&ik,&lk)) {
	 setException(PyExc_AttributeError,"sym_k should be set with a tuple (ik,lk)\n");
	 return -1;
      }
      if(ik!=0)
	 self->props->sym_k.ik=ik;
      if(lk!=0)
	 self->props->sym_k.lk=lk;
      return 0;
   } else if(strcmp(name,"istore_k")==0) {
      self->props->istore_k = PyInt_AsLong(val);
      return 0;
   } else if(strcmp(name,"meta_block_size")==0) {
      self->props->meta_block_size = PyInt_AsLong(val);
      return 0;
   }

   sprintf(errmsg,"It is not possible to set '%s' in the filecreationproperty instance\n",name);
   setException(PyExc_AttributeError,errmsg);
   
   return -1;
}

static PyObject* _getattr_compression(PyhlCompression* self, char* name)
{
    PyObject* res;

   if(strcmp(name,"type")==0) {
      return PyInt_FromLong(self->compr->type);
   } else if(strcmp(name,"level")==0) {
      return PyInt_FromLong(self->compr->level);
   } else if(strcmp(name,"szlib_mask")==0) {
      return PyInt_FromLong(self->compr->szlib_mask);
   } else if(strcmp(name,"szlib_px_per_block")==0) {
      return PyInt_FromLong(self->compr->szlib_px_per_block);
   } else if(strcmp(name,"H5_SZIP_CHIP_OPTION_MASK")==0) {
      return PyInt_FromLong(H5_SZIP_CHIP_OPTION_MASK);
   } else if(strcmp(name,"H5_SZIP_ALLOW_K13_OPTION_MASK")==0) {
      return PyInt_FromLong(H5_SZIP_ALLOW_K13_OPTION_MASK);
   } else if(strcmp(name,"H5_SZIP_EC_OPTION_MASK")==0) {
      return PyInt_FromLong(H5_SZIP_EC_OPTION_MASK);
   } else if(strcmp(name,"H5_SZIP_NN_OPTION_MASK")==0) {
      return PyInt_FromLong(H5_SZIP_NN_OPTION_MASK);
   }
   
   res = Py_FindMethod(compression_methods, (PyObject*) self, name);
   if (res)
      return res;
   
   PyErr_Clear();
   
   PyErr_SetString(PyExc_AttributeError, name);
   
   return NULL;   
}

static int _setattr_compression(PyhlCompression* self, char* name, PyObject* val)
{
   char errmsg[256];
   if(strcmp(name,"level")==0) {
      int tmpv;
      if(self->compr->type!=CT_ZLIB) {
	 setException(PyExc_AttributeError,"level is only usable when compression is of type COMPRESSION_ZLIB\n");
	 return -1;
      }
      tmpv = PyInt_AsLong(val);
      if(tmpv<0 || tmpv>9) {
	 setException(PyExc_AttributeError,"level must be specified as an integer between 0 and 9\n");
	 return -1;
      }
      self->compr->level = tmpv;
      return 0;
   } else if(strcmp(name,"szlib_mask")==0) {
      unsigned int compr_mask=H5_SZIP_CHIP_OPTION_MASK|H5_SZIP_ALLOW_K13_OPTION_MASK;
      unsigned int coding_mask=H5_SZIP_EC_OPTION_MASK|H5_SZIP_NN_OPTION_MASK;
      unsigned int allmask=compr_mask|coding_mask;
      unsigned int tmpv;
      if(self->compr->type!=CT_SZLIB) {
	 setException(PyExc_AttributeError,"szlib_mask is only usable when compression is of type COMPRESSION_SZLIB\n");
	 return -1;
      }
      /* Make sure valid values have been set for the mask*/
      tmpv = (unsigned int)PyInt_AsLong(val);
      /* Verify that both masks hasnt been set */
      if(tmpv&H5_SZIP_CHIP_OPTION_MASK && tmpv&H5_SZIP_ALLOW_K13_OPTION_MASK) {
	 setException(PyExc_AttributeError,"When setting szlib_mask only one of H5_SZIP_CHIP_OPTION_MASK and H5_SZIP_ALLOW_K13_OPTION_MASK may be used\n");
	 return -1;
      }
      if(tmpv&H5_SZIP_EC_OPTION_MASK && tmpv&H5_SZIP_NN_OPTION_MASK) {
	 setException(PyExc_AttributeError,"When setting szlib_mask only one of H5_SZIP_EC_OPTION_MASK and H5_SZIP_NN_OPTION_MASK mayb be used\n");
	 return -1;
      }
      /* Verify that at least one mask from each has been set */
      if((tmpv&compr_mask)==0 || (tmpv&coding_mask)==0) {
	 setException(PyExc_AttributeError,"the mask has to be set as a bitwise ORed value of H5_SZIP_CHIP_OPTION_MASK,H5_SZIP_ALLOW_K13_OPTION_MASK and H5_SZIP_ALLOW_K13_OPTION_MASK,H5_SZIP_EC_OPTION_MASK");
	 return -1;
      }
      if(tmpv&(!allmask)) {
	 setException(PyExc_AttributeError,"H5_SZIP_CHIP_OPTION_MASK,H5_SZIP_ALLOW_K13_OPTION_MASK,H5_SZIP_EC_OPTION_MASK"
		      ",H5_SZIP_NN_OPTION_MASK are the only flags that may be used when creating the bitmask\n");
	 return -1;
      }
      if(tmpv==0) {
	 setException(PyExc_AttributeError,"H5_SZIP_CHIP_OPTION_MASK,H5_SZIP_ALLOW_K13_OPTION_MASK,H5_SZIP_EC_OPTION_MASK"
		      ",H5_SZIP_NN_OPTION_MASK are the only flags that may be used when creating the bitmask\n");
	 return -1;
      }
      return 0;
   } else if(strcmp(name,"szlib_px_per_block")==0) {
      int tmpv;
      if(self->compr->type!=CT_SZLIB) {
	 setException(PyExc_AttributeError,"szlib_px_per_block is only usable when compression is of type COMPRESSION_SZLIB\n");
	 return -1;
      }
      // Make sure the value is even
      tmpv = PyInt_AsLong(val);
      if((tmpv%2)!=0) {
	 setException(PyExc_AttributeError,"szlib_px_per_block must be specified in even values, typically 8, 10, 16, and 32");
	 return -1;
      }
      self->compr->szlib_px_per_block=tmpv;
      Py_INCREF(Py_None);
      return 0;
   }

   sprintf(errmsg,"It is not possible to set '%s' in the compression instance\n",name);
   setException(PyExc_AttributeError,errmsg);   
   return -1;
}

statichere PyTypeObject Pyhl_Type = {
    PyObject_HEAD_INIT(NULL)
    0,				/*ob_size*/
    "PyhlCore",		/*tp_name*/
    sizeof(PyhlObject),	/*tp_size*/
    0,				/*tp_itemsize*/
    /* methods */
    (destructor)_dealloc,	/*tp_dealloc*/
    0,				/*tp_print*/
    (getattrfunc)_getattr,	/*tp_getattr*/
    0,				/*tp_setattr*/
    0,				/*tp_compare*/
    0,				/*tp_repr*/
    0,                          /*tp_as_number */
    0,
    0,                          /*tp_as_mapping */
    0                           /*tp_hash*/
};

statichere PyTypeObject PyhlNode_Type = {
    PyObject_HEAD_INIT(NULL)
    0,				/*ob_size*/
    "PyhlNodeCore",		/*tp_name*/
    sizeof(PyhlNode),	/*tp_size*/
    0,				/*tp_itemsize*/
    /* methods */
    (destructor)_dealloc_pyhlnode,/*tp_dealloc*/
    0,				/*tp_print*/
    (getattrfunc)_getattr_node,	/*tp_getattr*/
    0,				/*tp_setattr*/
    0,				/*tp_compare*/
    0,				/*tp_repr*/
    0,                          /*tp_as_number */
    0,
    0,                          /*tp_as_mapping */
    0                           /*tp_hash*/
};

statichere PyTypeObject PyhlFileCreationProperty_Type = {
    PyObject_HEAD_INIT(NULL)
    0,				/*ob_size*/
    "PyhlFileCreationProperty",	/*tp_name*/
    sizeof(PyhlFileCreationProperty),	/*tp_size*/
    0,				/*tp_itemsize*/
    /* methods */
    (destructor)_dealloc_pyhlfilecreationproperty,/*tp_dealloc*/
    0,				/*tp_print*/
    (getattrfunc)_getattr_filecreationproperty,	/*tp_getattr*/
    (setattrfunc)_setattr_filecreationproperty,	/*tp_setattr*/
    0,				/*tp_compare*/
    0,				/*tp_repr*/
    0,                          /*tp_as_number */
    0,
    0,                          /*tp_as_mapping */
    0                           /*tp_hash*/
};

statichere PyTypeObject PyhlCompression_Type = {
    PyObject_HEAD_INIT(NULL)
    0,				/*ob_size*/
    "PyhlCompression",	/*tp_name*/
    sizeof(PyhlCompression),	/*tp_size*/
    0,				/*tp_itemsize*/
    /* methods */
    (destructor)_dealloc_pyhlcompression,/*tp_dealloc*/
    0,				/*tp_print*/
    (getattrfunc)_getattr_compression,	/*tp_getattr*/
    (setattrfunc)_setattr_compression,	/*tp_setattr*/
    0,				/*tp_compare*/
    0,				/*tp_repr*/
    0,                          /*tp_as_number */
    0,
    0,                          /*tp_as_mapping */
    0                           /*tp_hash*/
};

static PyMethodDef functions[] = {
  {"nodelist", (PyCFunction)_pyhl_new_nodelist, 1},
  {"node", (PyCFunction)_pyhl_new_node, 1},
  {"filecreationproperty",(PyCFunction)_pyhl_new_filecreationproperty,1},
  {"compression",(PyCFunction)_pyhl_new_compression,1},
  {"read_nodelist",(PyCFunction)_pyhl_read_nodelist,1},
  {"is_file_hdf5",(PyCFunction)_pyhl_is_file_hdf5,1},
  {"show_hdf5errors",(PyCFunction)_pyhl_show_hdf5errors,1},
  {"show_hlhdferrors",(PyCFunction)_pyhl_show_hlhdferrors,1},
  {NULL,NULL} /*Sentinel*/
};

void init_pyhl(void)
{
   PyObject *m,*d;
   PyObject *tmp;
   Pyhl_Type.ob_type=&PyType_Type;
   PyhlNode_Type.ob_type=&PyType_Type;

   m=Py_InitModule("_pyhl",functions);
   d=PyModule_GetDict(m);

   ErrorObject = PyString_FromString("_pyhl.error");
   if(ErrorObject==NULL || PyDict_SetItemString(d,"error",ErrorObject)!=0) {
      Py_FatalError("Can't define _pyhl.error");
   }
   
   tmp = PyInt_FromLong(ATTRIBUTE_ID);
   PyDict_SetItemString(d,"ATTRIBUTE_ID",tmp);
   Py_XDECREF(tmp);

   tmp = PyInt_FromLong(GROUP_ID);
   PyDict_SetItemString(d,"GROUP_ID",tmp);
   Py_XDECREF(tmp);

   tmp = PyInt_FromLong(DATASET_ID);
   PyDict_SetItemString(d,"DATASET_ID",tmp);
   Py_XDECREF(tmp);

   tmp = PyInt_FromLong(TYPE_ID);
   PyDict_SetItemString(d,"TYPE_ID",tmp);
   Py_XDECREF(tmp);

   tmp = PyInt_FromLong(REFERENCE_ID);
   PyDict_SetItemString(d,"REFERENCE_ID",tmp);
   Py_XDECREF(tmp);   

   tmp = PyInt_FromLong(CT_ZLIB);
   PyDict_SetItemString(d,"COMPRESSION_ZLIB",tmp);
   Py_XDECREF(tmp);

   tmp = PyInt_FromLong(CT_SZLIB);
   PyDict_SetItemString(d,"COMPRESSION_SZLIB",tmp);
   Py_XDECREF(tmp);
 
   import_array(); /*To make sure I get access to Numeric*/
   /*Always have to do this*/
   initHlHdf();
   /*And this I just do to be able to get debugging info from hdf*/
   debugHlHdf(2);
}
