/* --------------------------------------------------------------------------
$Id: rave_info_type.c,v 1.3 2007/11/27 08:08:01 adybbroe Exp $
Program:        

Description:    Example python module for creating a compound type
		that can be modified from python and written to an
		HDF5 file.

Author(s):      Anders Henja

Copyright:	Swedish Meteorological and Hydrological Institute, 2000

History:
$Log: rave_info_type.c,v $
Revision 1.3  2007/11/27 08:08:01  adybbroe
Replaced PyMem_DEL with PyObject_Del, to be compatible with python 2.5
and beyond. From Python 2.5.1 release-notes
(http://www.python.org/download/releases/2.5.1/NEWS.txt):
PyMem_DEL no longer maps to PyObject_Free. It maps to the system
``free()`` now.  If memory is obtained via the ``PyObject_`` family,
it must be released via the ``PyObject_`` family, and likewise for the
``PyMem_`` family.

Revision 1.2  2003/01/15 11:57:19  ahenja
Added support for allowing both old and new numeric to be used.

Revision 1.1.1.1  2001/12/14 12:02:43  ahenja
First CVS version of hlhdf

Revision 1.3  2001/11/08 12:07:37  ahenja
Added area_extent for beeing able to test arrays in compound objects.

Revision 1.2  2000/03/22 08:59:06  ahenja
Updated with spell checking and header information.
	
-----------------------------------------------------------------------------*/
#include <Python.h>
#define HLHDF_PYMODULE_WITH_IMPORT_ARRAY
#include <arrayobject.h>

#include "pyhlhdf_common.h"
#include "read_vhlhdf.h"
#include "write_vhlhdf.h"

static PyObject *ErrorObject;

/* Some defines for error handling */
#define raiseException(type,msg)\
{PyErr_SetString(type,msg);return NULL;}

#define setException(type,msg)\
{PyErr_SetString(type,msg);}

typedef struct {
   int xsize;
   int ysize;
   double xscale;
   double yscale;
   double area_extent[4];
} RaveInfoStruct;

typedef struct {
   PyObject_HEAD /*Always have to be on top*/
   hid_t rave_info_hid;
} RaveInfoType;

typedef struct {
   PyObject_HEAD /*Always have to be on top*/
   RaveInfoStruct info;
} RaveInfoObject;


staticforward PyTypeObject RaveInfoType_Type;
#define RaveInfoType_Check(op) ((op)->ob_type == &RaveInfoType_Type)

staticforward PyTypeObject RaveInfoObject_Type;
#define RaveInfoObject_Check(op) ((op)->ob_type == &RaveInfoObject_Type)

static hid_t createRaveInfoType()
{
   hid_t type_id=-1;
   size_t dims[]={4};
   if((type_id=createCompoundType(sizeof(RaveInfoStruct)))<0)
      return -1;
   if(addAttributeToCompoundType_fmt(type_id,"xsize",HOFFSET(RaveInfoStruct,xsize),"int")<0)
      goto fail;
   if(addAttributeToCompoundType_fmt(type_id,"ysize",HOFFSET(RaveInfoStruct,ysize),"int")<0)
      goto fail;
   if(addAttributeToCompoundType_fmt(type_id,"xscale",HOFFSET(RaveInfoStruct,xscale),"double")<0)
      goto fail;
   if(addAttributeToCompoundType_fmt(type_id,"yscale",HOFFSET(RaveInfoStruct,yscale),"double")<0)
      goto fail;
   if(addArrayToCompoundType_fmt(type_id,"area_extent",HOFFSET(RaveInfoStruct,area_extent[0]),1,dims,"double")<0)
      goto fail;
   return type_id;
 fail:
   if(type_id>=0)
      H5Tclose(type_id);
   return -1;
}

static void _dealloc_type(RaveInfoType* rinfo)
{
   if(rinfo->rave_info_hid>=0) {
      disableErrorReporting();
      H5Tclose(rinfo->rave_info_hid);
      enableErrorReporting();
   }
   /*	PyMem_DEL(rinfo);
     PyMem_DEL no longer maps to PyObject_Free.
     It maps to the system ``free()`` now.  If memory
     is obtained via the ``PyObject_`` family, it must be released via the
     ``PyObject_`` family, and likewise for the ``PyMem_`` family.
     Adam Dybbroe, 2007-10-14
   */
   PyObject_Del(rinfo);
   
}

static PyObject* _rave_info_type_new_type(PyObject* self, PyObject* args)
{
   RaveInfoType* retv=NULL;
   retv=PyObject_NEW(RaveInfoType,&RaveInfoType_Type);
   if(!retv)
      return NULL;
   retv->rave_info_hid=-1;
   if((retv->rave_info_hid=createRaveInfoType())<0) {
      setException(PyExc_TypeError,"Could not create rave info type");
      _dealloc_type(retv);
      retv=NULL;
   }
   return (PyObject*)retv;
}

static void _dealloc_object(RaveInfoObject* robj)
{
   /*	PyMem_DEL(robj);
     PyMem_DEL no longer maps to PyObject_Free.
     It maps to the system ``free()`` now.  If memory
     is obtained via the ``PyObject_`` family, it must be released via the
     ``PyObject_`` family, and likewise for the ``PyMem_`` family.
     Adam Dybbroe, 2007-10-14
   */
   PyObject_Del(robj);
   
}

static PyObject* _rave_info_type_new_object(PyObject* self, PyObject* args)
{
   RaveInfoObject* retv=NULL;
   int i;
   retv=PyObject_NEW(RaveInfoObject,&RaveInfoObject_Type);
   if(!retv)
      return NULL;
   retv->info.xsize=0;
   retv->info.ysize=0;
   retv->info.xscale=0.0;
   retv->info.yscale=0.0;
   memset(retv->info.area_extent,0,sizeof(double)*4);
   return (PyObject*)retv;
}

static PyObject* _rave_info_type_hid(RaveInfoType* self, PyObject* args)
{
   return PyInt_FromLong(self->rave_info_hid);
}

static PyObject* _rave_info_type_size(RaveInfoType* self, PyObject* args)
{
   return PyInt_FromLong(sizeof(RaveInfoStruct));
}

static PyObject* _rave_info_object_fromstring(RaveInfoObject* self, PyObject* args)
{
   char* thestr=NULL;
   size_t len;
   char errbuf[256];
   if(!PyArg_ParseTuple(args,"s#",&thestr,&len))
      return NULL;
   if(len!=sizeof(RaveInfoStruct)) {
      sprintf(errbuf,"Length of data (%d) is not the same size as of RaveInfoStruct (%d)",
	      len,sizeof(RaveInfoStruct));
      setException(PyExc_TypeError,errbuf);
      return NULL;
   }
   memcpy(&self->info,thestr,sizeof(RaveInfoStruct));
   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject* _rave_info_object_tostring(RaveInfoObject* self, PyObject* args)
{
   return PyString_FromStringAndSize((char*)&self->info,sizeof(RaveInfoStruct));
}

static struct PyMethodDef type_methods[] =
{
   {"hid",(PyCFunction)_rave_info_type_hid,1},
   {"size",(PyCFunction)_rave_info_type_size,1},
   {NULL,NULL} /*Sentinel*/
};

static struct PyMethodDef object_methods[] =
{
   {"fromstring",(PyCFunction)_rave_info_object_fromstring,1},
   {"tostring",(PyCFunction)_rave_info_object_tostring,1},
   {"xsize",NULL},
   {"ysize",NULL},
   {"xscale",NULL},
   {"yscale",NULL},
   {"area_extent",NULL},
   {NULL,NULL} /*Sentinel*/
};

static PyObject* _getattr_type(RaveInfoType* self, char* name)
{
    PyObject* res;

    res = Py_FindMethod(type_methods, (PyObject*) self, name);
    if (res)
	return res;

    PyErr_Clear();

    PyErr_SetString(PyExc_AttributeError, name);

    return NULL;
}

static PyObject* _getattr_object(RaveInfoObject* self, char* name)
{
    PyObject* res;

    if(strcmp(name,"xsize")==0)
       return PyInt_FromLong(self->info.xsize);
    if(strcmp(name,"ysize")==0)
       return PyInt_FromLong(self->info.ysize);
    if(strcmp(name,"xscale")==0)
       return PyFloat_FromDouble(self->info.xscale);
    if(strcmp(name,"yscale")==0)
       return PyFloat_FromDouble(self->info.yscale);
    if(strcmp(name,"area_extent")==0) {
       return Py_BuildValue("(dddd)",self->info.area_extent[0],
			    self->info.area_extent[1],
			    self->info.area_extent[2],
			    self->info.area_extent[3]);
    }
    res = Py_FindMethod(object_methods, (PyObject*) self, name);
    if (res)
	return res;

    PyErr_Clear();

    PyErr_SetString(PyExc_AttributeError, name);

    return NULL;
}

static int _setattr_object(RaveInfoObject* self, char* name, PyObject* op)
{
   if(strcmp(name,"xsize")==0) {
      if(PyFloat_Check(op) || PyInt_Check(op)) {
	 self->info.xsize=PyInt_AsLong(op);
	 return 0;
      }
      setException(PyExc_ValueError,"Value not of type int");
      return -1;
   }
   if(strcmp(name,"ysize")==0) {
      if(PyFloat_Check(op) || PyInt_Check(op)) {
	 self->info.ysize=PyInt_AsLong(op);
	 return 0;
      }
      setException(PyExc_ValueError,"Value not of type int");
      return -1;
   }
   if(strcmp(name,"xscale")==0) {
      if(PyFloat_Check(op) || PyInt_Check(op)) {
	 self->info.xscale=PyFloat_AsDouble(op);
	 return 0;
      }
      setException(PyExc_ValueError,"Value not of type float");
      return -1;
   }
   if(strcmp(name,"yscale")==0) {
      if(PyFloat_Check(op) || PyInt_Check(op)) {
	 self->info.yscale=PyFloat_AsDouble(op);
	 return 0;
      }
      setException(PyExc_ValueError,"Value not of type float");
      return -1;
   }
   if(strcmp(name,"area_extent")==0) {
      if(PySequence_Check(op)) {
	 if(PyArg_ParseTuple(op,"dddd",&self->info.area_extent[0],
			     &self->info.area_extent[1],
			     &self->info.area_extent[2],
			     &self->info.area_extent[3])) {
	    return 0;
	 }
	 setException(PyExc_ValueError,"Area extent not set as tuple");
	 return -1;
      }
   }
   PyErr_SetString(PyExc_AttributeError,"Attribute does not exist or cannot be set");
   return -1;
}

statichere PyTypeObject RaveInfoType_Type = {
    PyObject_HEAD_INIT(NULL)
    0,				/*ob_size*/
    "RaveInfoTypeCore",		/*tp_name*/
    sizeof(RaveInfoType),	/*tp_size*/
    0,				/*tp_itemsize*/
    /* methods */
    (destructor)_dealloc_type,	/*tp_dealloc*/
    0,				/*tp_print*/
    (getattrfunc)_getattr_type,	/*tp_getattr*/
    0,				/*tp_setattr*/
    0,				/*tp_compare*/
    0,				/*tp_repr*/
    0,                          /*tp_as_number */
    0,
    0,                          /*tp_as_mapping */
    0                           /*tp_hash*/
};

statichere PyTypeObject RaveInfoObject_Type = {
    PyObject_HEAD_INIT(NULL)
    0,				/*ob_size*/
    "RaveInfoObjectCore",	/*tp_name*/
    sizeof(RaveInfoObject),	/*tp_size*/
    0,				/*tp_itemsize*/
    /* methods */
    (destructor)_dealloc_object,/*tp_dealloc*/
    0,				/*tp_print*/
    (getattrfunc)_getattr_object,/*tp_getattr*/
    (setattrfunc)_setattr_object,/*tp_setattr*/
    0,				/*tp_compare*/
    0,				/*tp_repr*/
    0,                          /*tp_as_number */
    0,
    0,                          /*tp_as_mapping */
    0                           /*tp_hash*/
};

static PyMethodDef functions[] = {
   {"type",(PyCFunction)_rave_info_type_new_type, 1},
   {"object",(PyCFunction)_rave_info_type_new_object, 1},
   {NULL,NULL}
};

void init_rave_info_type()
{
   PyObject *m,*d;
   PyObject *tmp;
   RaveInfoType_Type.ob_type=&PyType_Type;
   RaveInfoObject_Type.ob_type=&PyType_Type;

   m=Py_InitModule("_rave_info_type",functions);

   import_array(); /*To make sure I get access to Numeric*/
   /*Always have to do this*/
   initHlHdf();
   /*And this I just do to be able to get debugging info from hdf*/
   debugHlHdf(2);
}
